# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" RTI quicklook.
"""
import logging

import numpy as np

from argos.qt import QtGui, QtWidgets
from argos.repo.detailpanes import DetailBasePane
from argos.widgets.constants import MONO_FONT, FONT_SIZE

logger = logging.getLogger(__name__)


class QuickLookPane(DetailBasePane):
    """ Shows a string representation of the RTI contents.
    """
    _label = "Quick Look"

    def __init__(self, repoTreeView, parent=None):
        super(QuickLookPane, self).__init__(repoTreeView, parent=parent)

        self._currentRti = None

        self.editor = QtWidgets.QPlainTextEdit()
        self.editor.setReadOnly(True)
        self.editor.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.editor.setFont(QtGui.QFont(MONO_FONT, FONT_SIZE))

        self.contentsLayout.addWidget(self.editor)


    def _drawContents(self, currentRti=None):
        """ Draws the attributes of the currentRTI
        """
        self._currentRti = currentRti

        if self._currentRti is None:
            self.editor.clear()
        else:
            editorCharWidth =  self.editor.width() / self.editor.fontMetrics().averageCharWidth()
            oldLineWidth = np.get_printoptions()['linewidth']
            np.set_printoptions(linewidth=editorCharWidth)
            try:
                self.editor.setPlainText(self._currentRti.quickLook(editorCharWidth))
            finally:
                np.set_printoptions(linewidth=oldLineWidth)


    def resizeEvent(self, event):
        """ Called when the panel is resized. Will update the line length of the editor.
        """
        self.repoItemChanged(self._currentRti) # call repoItemChanged so it handles exceptions
        super(QuickLookPane, self).resizeEvent(event)




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.


""" Editable table for data store
"""

import logging
import os.path

from argos.utils.cls import type_name, check_class

from argos.info import icons_directory
from argos.reg.tabmodel import BaseTableModel
from argos.qt import QtCore, QtGui, QtWidgets, Qt
from argos.qt.togglecolumn import ToggleColumnTableView


logger = logging.getLogger(__name__)


class BaseTableView(ToggleColumnTableView):
    """ Editable QTableView that shows the contents of a BaseTableModel.
    """
    def __init__(self, model=None, parent=None):
        """ Constructor

            :param BaseTableModelmodel: a RegistryTableModel that maps the regItems
            :param QWidget parent: the parent widget
        """
        super(BaseTableView, self).__init__(parent)

        check_class(model, BaseTableModel)
        self.setModel(model)

        self.setTextElideMode(QtCore.Qt.ElideMiddle) # Does not work nicely when editing cells.
        self.setAlternatingRowColors(True)
        self.setShowGrid(False)
        self.setSortingEnabled(False)
        self.setTabKeyNavigation(False)
        self.setWordWrap(False)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        Qiv = QtWidgets.QAbstractItemView
        #self.setEditTriggers(Qiv.DoubleClicked | Qiv.SelectedClicked | Qiv.EditKeyPressed)
        self.setEditTriggers(Qiv.NoEditTriggers)
        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

        self.verHeader = self.verticalHeader()
        self.verHeader.setSectionsMovable(False)
        self.verHeader.hide()

        self.horHeader = self.horizontalHeader()
        self.horHeader.setDefaultAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        self.horHeader.setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.horHeader.setStretchLastSection(False)

        for col, canStretch in enumerate(model.store.canStretchPerColumn):
            if canStretch:
                self.horHeader.setSectionResizeMode(col, QtWidgets.QHeaderView.Stretch)
            else:
                self.horHeader.setSectionResizeMode(col, QtWidgets.QHeaderView.ResizeToContents)


    def getCurrentItem(self):
        """ Returns the item of the selected row, or None if none is selected
        """
        curIdx = self.currentIndex()
        return self.model().itemFromIndex(curIdx)


    def setCurrentCell(self, row, col=0):
        """ Sets the current row and column.
        """
        cellIdx = self.model().index(row, col)
        if not cellIdx.isValid():
            logger.warning("Can't set (row = {}, col = {}) in table".format(row, col))
        else:
            logger.debug("Setting cellIdx (row = {}, col = {}) in table".format(row, col))
        self.setCurrentIndex(cellIdx)



class TableEditWidget(QtWidgets.QWidget):
    """ Widget that contains a table plus buttons to update it

    """
    def __init__(self, tableModel=None, parent=None):
        """ Constructor.
        """
        super(TableEditWidget, self).__init__(parent=parent)

        self.setFocusPolicy(Qt.NoFocus)

        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.mainLayout)

        self.tableView = BaseTableView(tableModel)
        self.mainLayout.addWidget(self.tableView)

        buttonLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.addLayout(buttonLayout)

        iconDir = icons_directory()
        iconSize = QtCore.QSize(20, 20)

        self.addButton = QtWidgets.QPushButton()
        self.addButton.setToolTip("Add new row.")
        self.addButton.setIcon(QtGui.QIcon(os.path.join(iconDir, 'plus-sign-l.svg')))
        self.addButton.setIconSize(iconSize)
        self.addButton.clicked.connect(self.addRow)
        buttonLayout.addWidget(self.addButton)

        self.removeButton = QtWidgets.QPushButton()
        self.removeButton.setToolTip("Remove row.")
        self.removeButton.setIcon(QtGui.QIcon(os.path.join(iconDir, 'minus-sign-l.svg')))
        self.removeButton.setIconSize(iconSize)
        self.removeButton.clicked.connect(self.removeRow)
        buttonLayout.addWidget(self.removeButton)
        buttonLayout.addSpacing(25)

        self.moveUpButton = QtWidgets.QPushButton()
        self.moveUpButton.setToolTip("Move row up")
        self.moveUpButton.setIcon(QtGui.QIcon(os.path.join(iconDir, 'circle-arrow-up-l.svg')))
        self.moveUpButton.setIconSize(iconSize)
        self.moveUpButton.clicked.connect(lambda: self.moveRow(-1))
        buttonLayout.addWidget(self.moveUpButton)

        self.moveDownButton = QtWidgets.QPushButton()
        self.moveDownButton.setToolTip("Move row down")
        self.moveDownButton.setIcon(QtGui.QIcon(os.path.join(iconDir, 'circle-arrow-down-l.svg')))
        self.moveDownButton.setIconSize(iconSize)
        self.moveDownButton.clicked.connect(lambda: self.moveRow(+1))
        buttonLayout.addWidget(self.moveDownButton)

        buttonLayout.addStretch()

        self.tableView.selectionModel().currentChanged.connect(self.onCurrentChanged)
        self.tableView.setFocus(Qt.NoFocusReason)
        self.updateWidgets()


    def addRow(self):
        """ Adds an empty row in the plugin table
        """
        model = self.tableView.model()
        # curIdx = self.tableView.getSectionCurrentIndex()
        curIdx = self.tableView.currentIndex()
        curRow, curCol = curIdx.row(), curIdx.column()
        curRow = model.rowCount() if curRow < 0 else curRow # append at the end if non selected
        curCol = 0 if curCol < 0 else curCol
        model.insertItem(model.createItem(), row=curRow)

        self.tableView.setCurrentCell(curRow, curCol)


    def removeRow(self):
        """ Removes the currently selected row
        """
        model = self.tableView.model()
        # curIdx = self.tableView.getSectionCurrentIndex()
        curIdx = self.tableView.currentIndex()
        curRow = curIdx.row()
        if curRow < 0:
            logger.warning("No row selected so no row removed.")
            return

        model.popItemAtRow(curRow)
        self.tableView.setCurrentCell(min(curRow, model.rowCount() - 1), curIdx.column())


    def moveRow(self, number):
        """ Moves the currently selected row a with a number of positions
        """
        #curIdx = self.tableView.getSectionCurrentIndex()
        model = self.tableView.model()
        curIdx = self.tableView.currentIndex()
        curRow = curIdx.row()
        if curRow < 0:
            logger.warning("No row selected so no row moved.")
            return

        model.moveItem(curRow, curRow + number)
        newRow = max(0, min(curRow + number, model.rowCount() - 1) ) # clip
        self.tableView.setCurrentCell(newRow, curIdx.column())


    def onCurrentChanged(self, _curIdx, _prefIdx):
        """ Called when the current table index changes
        """
        self.updateWidgets()


    def updateWidgets(self):
        """ Enables/disables widgets according to the selected row
        """
        row = self.tableView.currentIndex().row()
        numRows = self.tableView.model().rowCount()

        self.addButton.setEnabled(True)
        self.removeButton.setEnabled(0 <= row < numRows)
        self.moveUpButton.setEnabled(1 <= row < numRows)
        self.moveDownButton.setEnabled(0 <= row < numRows-1)



def main():
    """ Test classes
    """
    import sys
    from pprint import pprint

    from PyQt5 import QtWidgets

    from argos import info
    from argos.utils.logs import make_log_format
    from argos.qt.misc import handleException
    from argos.reg.tabmodel import BaseTableModel, BaseItemStore, BaseItem

    info.DEBUGGING = True
    sys.excepthook = handleException

    logging.basicConfig(level="DEBUG", format=make_log_format())


    app = QtWidgets.QApplication([])

    store = BaseItemStore()


    model = BaseTableModel(store)

    window = TableEditWidget(tableModel=model)
    window.show()

    exitCode = app.exec_()

    print("edited store")
    pprint(store.marshall())





if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

"""
    OpenInspectorDialog dialog window that lets the user pick a new inspector.
"""
from __future__ import print_function

import logging

from argos.qt import Qt, QtCore, QtWidgets
from argos.inspector.registry import InspectorRegItem
from argos.utils.cls import check_class
from argos.widgets.pluginsdialog import RegistryTab


logger = logging.getLogger(__name__)


# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201

class OpenInspectorDialog(QtWidgets.QDialog):
    """ Dialog window that allows the user to open an inspector plugins.

        SIDE EFFECT: will try to import all underlying inspector classes.
            This is done so that help and number of dimensions can be displayed.
    """

    def __init__(self, registry, onlyShowImported=False, parent=None):
        """ Constructor
        """
        super(OpenInspectorDialog, self).__init__(parent=parent)

        self._registry = registry

        self.setWindowTitle('Open Inspector')
        self.setModal(True)
        layout = QtWidgets.QVBoxLayout(self)

        attrNames = ['name', 'library', 'nDims']
        headerSizes = [250, 250, None]

        self.inspectorTab = RegistryTab(registry, attrNames=attrNames, headerSizes=headerSizes,
                                        onlyShowImported=onlyShowImported)
        self.inspectorTab.tableView.sortByColumn(1, Qt.AscendingOrder) # sort by library
        layout.addWidget(self.inspectorTab)

        # Buttons
        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok |
                                           QtWidgets.QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        layout.addWidget(buttonBox)

        # Double clicking is equivalent to selecting it and clicking Ok.
        self.inspectorTab.tableView.doubleClicked.connect(self.accept)

        self.resize(QtCore.QSize(800, 600))

        #self.inspectorTab.tryImportAllPlugins()


    def getCurrentInspectorRegItem(self):
        """ Returns the inspector that is currently selected in the table.
            Can return None if there is no data in the table
        """
        return self.inspectorTab.getCurrentRegItem()


    def setCurrentInspectorRegItem(self, regItem):
        """ Sets the current inspector given an InspectorRegItem
        """
        check_class(regItem, InspectorRegItem, allow_none=True)
        self.inspectorTab.setCurrentRegItem(regItem)






""" Harmonize PyQt5 and PySide2 bindings

    If the QT_API environment variable is set to pyside2 or pyqt5 that library is imported.
    Else it looks if one of those two is already imported.

    Then it will try first PyQt5 and then PySide2.

    Note that PySide2 at the moment does not fully work yet. There are problems with the color
    picker. Also (for now) is not officially supported! There are already enough dependencies
    that can vary (Python 2 & 3, Windows & Linux & OS-X, etc) so I don't want to support even more
    combinations.
"""
import logging
import os
import sys

logger = logging.getLogger(__name__)

API_PYSIDE2 = 'pyside2'
API_PYQT5 = 'pyqt5'
ALL_API = [API_PYQT5, API_PYSIDE2]

# Determine from environment variable
QT_API_NAME = os.environ.get('QT_API')

# Discard unsupported bindings (e.g. pyqt4)
if QT_API_NAME not in ALL_API:
    QT_API_NAME = None

# Else see if module is already imported.
if QT_API_NAME is None:
    if 'PyQt5' in sys.modules:
        QT_API_NAME = API_PYQT5

if QT_API_NAME is None:
    if 'PySide2' in sys.modules:
        QT_API_NAME = API_PYSIDE2


# Finally, try the modules.
if QT_API_NAME is None:
    try:
        import PyQt5
    except ModuleNotFoundError as ex:
        logger.debug("Tried but failed to import PyQt5: {}".format(ex))
        pass
    else:
        QT_API_NAME = API_PYQT5

if QT_API_NAME is None:
    try:
        import PySide2
    except ModuleNotFoundError as ex:
        logger.debug("Tried but failed to import PySide2: {}".format(ex))
    else:
        QT_API_NAME = API_PYSIDE2


# Do imports.

if QT_API_NAME == API_PYQT5:
    from PyQt5 import QtCore, QtGui, QtWidgets, QtSvg
    from PyQt5.QtCore import Qt, QUrl
    from PyQt5.QtCore import pyqtSignal as QtSignal
    from PyQt5.QtCore import pyqtSlot as QtSlot
    from PyQt5.Qt import PYQT_VERSION_STR as PYQT_VERSION
    from PyQt5.Qt import QT_VERSION_STR as QT_VERSION

elif QT_API_NAME == API_PYSIDE2:

    from PySide2 import QtCore, QtGui, QtWidgets, QtSvg
    from PySide2.QtCore import Qt, QUrl
    from PySide2.QtCore import Signal as QtSignal
    from PySide2.QtCore import Slot as QtSlot
    from PySide2 import __version__ as PYQT_VERSION
    from PySide2.QtCore import __version__ as QT_VERSION

else:

    raise RuntimeError(
        "Unable to import one of the Qt bindings: {}. The specified binding "
        "(QT_API environment variable) is: {!r})".format(ALL_API, QT_API_NAME))





# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Defines a global Inspector registry to register inspector plugins.
"""

import logging

from argos.info import DEBUGGING
from argos.reg.basereg import BaseRegItem, BaseRegistry, RegType

logger = logging.getLogger(__name__)

DEFAULT_INSPECTOR = 'Table'

class InspectorRegItem(BaseRegItem):
    """ Class to keep track of a registered Inspector.
        Has a create() method that functions as an Inspector factory.
    """
    FIELDS  = BaseRegItem.FIELDS[:1] + ['shortCut'] + BaseRegItem.FIELDS[1:]
    TYPES   = BaseRegItem.TYPES[:1] + [RegType.ShortCut] + BaseRegItem.TYPES[1:]
    LABELS  = BaseRegItem.LABELS[:1] + ['Shortcut'] + BaseRegItem.LABELS[1:]
    STRETCH = BaseRegItem.STRETCH[:1] + [False] + BaseRegItem.STRETCH[1:]


    def __init__(self, name='', absClassName='', pythonPath='', shortCut=''):
        """ Constructor. See the ClassRegItem class doc string for the parameter help.
        """
        super(InspectorRegItem, self).__init__(name=name, absClassName=absClassName,
                                               pythonPath=pythonPath)
        self._data['shortCut'] = shortCut


    @property
    def shortCut(self):
        """ Keyboard short cut
        """
        return self._data['shortCut']


    @property
    def axesNames(self):
        """ The axes names of the inspector.
        """
        return [] if self.cls is None else self.cls.axesNames()

    @property
    def nDims(self):
        """ The number of axes of this inspector.
        """
        return len(self.axesNames)


    def create(self, collector, tryImport=True):
        """ Creates an inspector of the registered and passes the collector to the constructor.
            Tries to import the class if tryImport is True.
            Raises ImportError if the class could not be imported.
        """
        cls = self.getClass(tryImport=tryImport)
        if not self.successfullyImported:
            raise ImportError("Class not successfully imported: {}".format(self.exception))
        return cls(collector)



class InspectorRegistry(BaseRegistry):
    """ Class that maintains the collection of registered inspector classes.
        See the base class documentation for more info.
    """
    ITEM_CLASS = InspectorRegItem

    def __init__(self):
        """ Constructor
        """
        super(InspectorRegistry, self).__init__()


    @property
    def registryName(self):
        """ Human readable name for this registry.
        """
        return "Inspector"


    def getDefaultItems(self):
        """ Returns a list with the default plugins in the inspector registry.
        """
        plugins = [
            InspectorRegItem('Line Plot',
                             'argos.inspector.pgplugins.lineplot1d.PgLinePlot1d',
                             shortCut='Ctrl+1'),
            InspectorRegItem('Image Plot',
                             'argos.inspector.pgplugins.imageplot2d.PgImagePlot2d',
                             shortCut='Ctrl+2'),
            InspectorRegItem(DEFAULT_INSPECTOR,
                             'argos.inspector.qtplugins.table.TableInspector',
                             shortCut='Ctrl+3'),
            InspectorRegItem('Text',
                             'argos.inspector.qtplugins.text.TextInspector',
                             shortCut='Ctrl+4'),
            ]
        if DEBUGGING:
            plugins.append(InspectorRegItem('Image Plot (Old)',
                                            'argos.inspector.pgplugins.old_imageplot2d.PgImagePlot2d'))
            plugins.append(InspectorRegItem('Debug Inspector',
                                            'argos.inspector.debug.DebugInspector'))
        return plugins



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Inspector Selection Pane
"""
import logging

from argos.qt import QtWidgets, QtSlot, Qt

from argos.inspector.registry import InspectorRegItem
from argos.widgets.constants import DOCK_MARGIN, DOCK_SPACING
from argos.widgets.misc import setWidgetSizePolicy

logger = logging.getLogger(__name__)

NO_INSPECTOR_LABEL = '<No Inspector>'


class InspectorSelectionPane(QtWidgets.QFrame):
    """ Shows the attributes of the selected repo tree item
    """

    def __init__(self, inspectorActionGroup, parent=None):
        super(InspectorSelectionPane, self).__init__(parent=parent)

        # self.setFrameShape(QtWidgets.QFrame.Box)
        self.menuButton = QtWidgets.QPushButton(NO_INSPECTOR_LABEL)
        self.menuButton.setMinimumWidth(10)

        inspectorMenu = QtWidgets.QMenu("Change Inspector", parent=self.menuButton)
        for action in inspectorActionGroup.actions():
            inspectorMenu.addAction(action)
        self.menuButton.setMenu(inspectorMenu)

        self.messageLabel = QtWidgets.QLabel("")
        self.messageLabel.setObjectName("message_label")
        self.messageLabel.setFrameStyle(QtWidgets.QFrame.Panel)
        self.messageLabel.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.messageLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)

        self.mainLayout = QtWidgets.QHBoxLayout()
        self.mainLayout.setContentsMargins(DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN, DOCK_MARGIN)
        self.setLayout(self.mainLayout)

        self.mainLayout.addWidget(self.menuButton, stretch=0)
        self.mainLayout.addWidget(self.messageLabel, stretch=1)

        setWidgetSizePolicy(self.menuButton, hor=QtWidgets.QSizePolicy.Minimum)
        setWidgetSizePolicy(self.messageLabel, hor=QtWidgets.QSizePolicy.Ignored)
        setWidgetSizePolicy(
            self, hor=QtWidgets.QSizePolicy.MinimumExpanding, ver=QtWidgets.QSizePolicy.Fixed)


    def showMessage(self, msg):
        """ Sets the message label to a message text
        """
        logger.debug("Show message to user: '{}'".format(msg))
        self.messageLabel.setText(msg)
        self.messageLabel.setToolTip(msg)


    @QtSlot(InspectorRegItem)
    def updateFromInspectorRegItem(self, inspectorRegItem):
        """ Updates the label from the full name of the InspectorRegItem
        """
        label = NO_INSPECTOR_LABEL if inspectorRegItem is None else inspectorRegItem.name
        self.menuButton.setText(label)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Contains TableInspector and TableInspectorModel
"""
import logging
import numbers

import numpy as np

from argos.external import six
from argos.collect.collector import FAKE_DIM_NAME
from argos.config.boolcti import BoolCti
from argos.config.choicecti import ChoiceCti
from argos.config.groupcti import GroupCti, MainGroupCti
from argos.config.intcti import IntCti
from argos.config.qtctis import FontCti, ColorCti
from argos.info import DEBUGGING
from argos.inspector.abstract import AbstractInspector
from argos.qt import Qt, QtCore, QtGui, QtWidgets
from argos.widgets.constants import MONO_FONT, FONT_SIZE
from argos.widgets.argostableview import ArgosTableView
from argos.utils.cls import check_class, check_is_a_string
from argos.utils.cls import to_string, is_an_array
from argos.utils.misc import is_quoted

logger = logging.getLogger(__name__)

RESET_HEADERS_AT_SIZE = 10000   # If header has more elements, the headers size CTIs are disabled.
OPTIMIZE_RESIZE_AT_SIZE = 1000  # If the table has more elements, only the current column/row are
                                # resized and the others take the same size

ALIGN_SMART = -1  # Use right alignment for numbers and left alignment for everything else.

def resizeAllSections(header, sectionSize):
    """ Sets all sections (columns or rows) of a header to the same section size.

        :param header: a QHeaderView
        :param sectionSize: the new size of the header section in pixels
    """
    for idx in range(header.length()):
        header.resizeSection(idx, sectionSize)


def makeReplacementField(formatSpec, altFormatSpec='', testValue=None):
    """ Prepends a colon and wraps the formatSpec in curly braces to yield a replacement field.

        The format specification is part of a replacement field, which can be used in new-style
        string formatting. See:
            https://docs.python.org/3/library/string.html#format-string-syntax
            https://docs.python.org/3/library/string.html#format-specification-mini-language

        If the formatSpec does not contain a colon or exclamation mark, a colon is prepended.

        If the formatSpec starts and end in quotes (single or double) only the quotes are removed,
        no curly braces or colon charactes are added. This allows users to define a format spec.

        :param formatSpec: e.g. '5.2f' will return '{:5.2f}'
        :param altFormatSpec: alternative that will be used if the formatSpec evaluates to False
        :param testValue: if not None, result.format(testValue) will be evaluated as a test.
        :return: string
    """
    check_is_a_string(formatSpec)
    check_is_a_string(altFormatSpec)
    fmt = altFormatSpec if not formatSpec else formatSpec

    if is_quoted(fmt):
        fmt = fmt[1:-1] # remove quotes
    else:
        if fmt and ':' not in fmt and '!' not in fmt:
            fmt = ':' + fmt
        fmt = '{' + fmt + '}'

    # Test resulting replacement field
    if testValue is not None:
        try:
            _dummy = fmt.format(testValue)
        except Exception:
            msg = ("Format specifier failed: replacement-field={!r}, test-value={!r}"
                   .format(fmt, testValue))
            logger.error(msg)
            raise ValueError(msg)

    logger.debug("Resulting replacement field: {!r}".format(fmt))
    return fmt



class TableInspectorCti(MainGroupCti):
    """ Configuration tree item for a TableInspector
    """
    def __init__(self, tableInspector, nodeName):

        super(TableInspectorCti, self).__init__(nodeName)

        check_class(tableInspector, TableInspector)
        self.tableInspector = tableInspector

        # The defaultRowHeightCti and defaultColWidthCti are initialized with -1; they will get
        # the actual values in the TableInspector constructor.
        self.autoRowHeightCti = self.insertChild(BoolCti("auto row heights", False,
                                                        childrenDisabledValue=True))
        self.defaultRowHeightCti = self.autoRowHeightCti.insertChild(
            IntCti("row height", -1, minValue=20, maxValue=500, stepSize=5))

        self.autoColWidthCti = self.insertChild(BoolCti("auto column widths", False,
                                                        childrenDisabledValue=True))
        self.defaultColWidthCti = self.autoColWidthCti.insertChild(
            IntCti("column width", -1, minValue=20, maxValue=500, stepSize=5))

        self.insertChild(BoolCti("separate fields", True))
        self.insertChild(BoolCti("word wrap", False))

        self.encodingCti = self.insertChild(
            ChoiceCti('encoding', editable=True,
                      configValues=['utf-8', 'ascii', 'latin-1', 'windows-1252']))

        fmtCti = self.insertChild(GroupCti("format specifiers"))

        # # Use '!r' as default for Python 2. This will convert the floats with repr(), which is
        # # necessary because str() or an empty format string will only print 2 decimals behind the
        # # point. In Python 3 this is not necessary: all relevant decimals are printed.
        #numDefaultIndex = 6 if six.PY2 else 0
        numDefaultIndex = 7

        self.strFormatCti = fmtCti.insertChild(
            ChoiceCti("strings", 0, editable=True, completer=None,
                      configValues=['', 's', '!r', '!a',
                                    '10.10s', '_<15s', '_>15s', "'str: {}'"]))

        self.intFormatCti = fmtCti.insertChild(
            ChoiceCti("integers", 0, editable=True, completer=None,
                      configValues=['', 'd', 'n', 'c', '#b', '#o', '#x', '!r',
                                    '8d', '#8.4g', '_<10', '_>10', "'int: {}'"]))

        self.numFormatCti = fmtCti.insertChild(
            ChoiceCti("other numbers", numDefaultIndex, editable=True, completer=None,
                      configValues=['', 'f', 'g', 'n', '%', '!r',
                                    '8.3e', '#8.4g', '_<15', '_>15', "'num: {}'"]))

        self.otherFormatCti = fmtCti.insertChild(
            ChoiceCti("other types", 0, editable=True, completer=None,
                      configValues=['', '!s', '!r', "'other: {}'"]))

        self.maskFormatCti = fmtCti.insertChild(
            ChoiceCti("missing data", 2, editable=True, completer=None,
                      configValues=['', "' '", "'--'", "'<masked>'", '!r']))

        self.fontCti = self.insertChild(FontCti(self.tableInspector, "font",
                                                defaultData=QtGui.QFont(MONO_FONT, FONT_SIZE)))

        self.dataColorCti = self.insertChild(
            ColorCti('text color', QtGui.QColor('#000000')))

        self.missingColorCti = self.insertChild(
            ColorCti('missing data color', QtGui.QColor('#B0B0B0')))

        self.horAlignCti = self.insertChild(
            ChoiceCti('horizontal alignment',
                      configValues=[ALIGN_SMART, Qt.AlignLeft, Qt.AlignHCenter, Qt.AlignRight],
                      displayValues=['Type-dependent', 'Left', 'Center', 'Right']))
                      # Qt.AlignJustify not included, it doesn't seem to do anything,

        self.verAlignCti = self.insertChild(
            ChoiceCti('vertical alignment', defaultData=1,
                      configValues=[Qt.AlignTop, Qt.AlignVCenter, Qt.AlignBottom],
                      displayValues=['Top', 'Center', 'Bottom']))

        # Per pixel scrolling works better for large cells (e.g. containing XML strings).
        self.insertChild(ChoiceCti("scroll", displayValues=["Per cell", "Per pixel"],
                                   configValues=[QtWidgets.QAbstractItemView.ScrollPerItem,
                                                 QtWidgets.QAbstractItemView.ScrollPerPixel]))


    def _refreshNodeFromTarget(self):
        """ Refreshes the TableInspectorCti from the TableInspector target it monitors.

            Disables auto-sizing of the header sizes for very large headers (> 10000 elements).
            Otherwise the resizing may take to long and the program will hang.
        """
        tableModel = self.tableInspector.model

        # Disable row height and column with settings for large headers (too slow otherwise)
        if tableModel.rowCount() >= RESET_HEADERS_AT_SIZE:
            self.autoRowHeightCti.data = False
            self.model.emitDataChanged(self.autoRowHeightCti)
            self.autoRowHeightCti.enable = False
        else:
            self.autoRowHeightCti.enable = True

        if tableModel.columnCount() >= RESET_HEADERS_AT_SIZE:
            self.autoColWidthCti.data  = False
            self.model.emitDataChanged(self.autoColWidthCti)
            self.autoColWidthCti.enable = False
        else:
            self.autoColWidthCti.enable = True


    def resetRangesToDefault(self):
        """ Resets columns auto height and width
        """
        self.defaultColWidthCti.resetToDefault()
        self.defaultRowHeightCti.resetToDefault()
        self.autoColWidthCti.resetToDefault()
        self.autoRowHeightCti.resetToDefault()



class TableInspector(AbstractInspector):
    """ Displays the array data in a table.
    """
    def __init__(self, collector, parent=None):

        super(TableInspector, self).__init__(collector, parent=parent)

        self.model = TableInspectorModel(parent=self)
        self.tableView = ArgosTableView()
        self.contentsLayout.addWidget(self.tableView)
        self.tableView.setModel(self.model)
        self.tableView.setSelectionMode(QtWidgets.QTableView.ContiguousSelection)

        horHeader = self.tableView.horizontalHeader()
        verHeader = self.tableView.verticalHeader()
        horHeader.setCascadingSectionResizes(False)
        verHeader.setCascadingSectionResizes(False)

        self._config = TableInspectorCti(tableInspector=self, nodeName='table')

        if self.config.defaultRowHeightCti.configValue < 0: # If not yet initialized
            self.config.defaultRowHeightCti.data = verHeader.defaultSectionSize()
            self.config.defaultRowHeightCti.defaultData = verHeader.defaultSectionSize()

        if self.config.defaultColWidthCti.configValue < 0: # If not yet initialized
            self.config.defaultColWidthCti.data = horHeader.defaultSectionSize()
            self.config.defaultColWidthCti.defaultData = horHeader.defaultSectionSize()


    @classmethod
    def axesNames(cls):
        """ The names of the axes that this inspector visualizes.
            See the parent class documentation for a more detailed explanation.
        """
        return tuple(['Y', 'X'])


    def getFont(self):
        """ Returns the font of the table model. Can be a QFont or None if no font is set.
        """
        return self.model.getFont()


    def setFont(self, font):
        """ Sets the font of the table model. Can be a QFont or None if no font is set.
        """
        return self.model.setFont(font)


    def _clearContents(self):
        """ Clears the inspector widget when no valid input is available.
        """
        pass


    def _drawContents(self, reason=None, initiator=None):
        """ Draws the table contents from the sliced array of the collected repo tree item.

            See AbstractInspector.updateContents for the reason and initiator description.
        """
        logger.debug("TableInspector._drawContents: {}".format(self))

        # If auto-reset is true, reset config complete or partially, depending on the mode.
        if self._resetRequired(reason, initiator):
            self.resetConfig()

        oldTableIndex = self.tableView.currentIndex()
        if oldTableIndex.isValid():
            selectionWasValid = True # Defensive programming, keep old value just in case
            oldRow = oldTableIndex.row()
            oldCol = oldTableIndex.column()
        else:
            selectionWasValid = False
            oldRow = 0
            oldCol = 0

        # Temporarily set header sizes fixed when populating the model (it may be slow otherwise)
        verHeader = self.tableView.verticalHeader()
        horHeader = self.tableView.horizontalHeader()
        verHeader.setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
        horHeader.setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        self.model.updateState(self.collector.getSlicedArray(),
                               self.collector.rtiInfo,
                               self.configValue('separate fields'))

        self.model.encoding = self.config.encodingCti.configValue
        self.model.horAlignment = self.config.horAlignCti.configValue
        self.model.verAlignment = self.config.verAlignCti.configValue
        self.model.dataColor = self.config.dataColorCti.configValue
        self.model.missingColor = self.config.missingColorCti.configValue

        self.model.strFormat   = makeReplacementField(self.config.strFormatCti.configValue,
                                                      testValue='my_string')
        self.model.intFormat   = makeReplacementField(self.config.intFormatCti.configValue,
                                                      testValue=0)
        self.model.numFormat   = makeReplacementField(self.config.numFormatCti.configValue,
                                                      testValue=0.0)
        self.model.otherFormat = makeReplacementField(self.config.otherFormatCti.configValue,
                                                      testValue=None)
        self.model.maskFormat  = makeReplacementField(self.config.maskFormatCti.configValue,
                                                      testValue=None)

        scrollMode = self.configValue("scroll")
        self.tableView.setHorizontalScrollMode(scrollMode)
        self.tableView.setVerticalScrollMode(scrollMode)
        self.tableView.setWordWrap(self.configValue('word wrap'))

        # Update the model font from the font config item (will call self.setFont)
        self.config.updateTarget()

        if (self.config.autoRowHeightCti.configValue
            and self.model.rowCount() < RESET_HEADERS_AT_SIZE):

            numCells = self.model.rowCount() * self.model.columnCount()
            if numCells <= OPTIMIZE_RESIZE_AT_SIZE:
                logger.debug("Setting vertical resize mode to ResizeToContents")
                verHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
            else:
                # ResizeToContents can be very slow because it gets all rows.
                # Work around: resize only first row and columns and apply this to all rows/cols
                logger.warning("Performance work around: for tables with more than 1000 cells " +
                               "only the current row is resized and the others use that size.")
                verHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
                self.tableView.resizeRowToContents(oldRow)
                resizeAllSections(verHeader, self.tableView.rowHeight(oldRow))
        else:
            logger.debug("Setting vertical resize mode to Interactive and reset header")
            verHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
            verHeader.setDefaultSectionSize(self.config.defaultRowHeightCti.configValue)
            verHeader.reset()

        if (self.config.autoColWidthCti.configValue
            and self.model.columnCount() < RESET_HEADERS_AT_SIZE):

            numCells = self.model.rowCount() * self.model.columnCount()
            if numCells <= OPTIMIZE_RESIZE_AT_SIZE:
                logger.debug("setting horizontal resize mode to ResizeToContents")
                horHeader.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
            else:
                # ResizeToContents can be very slow because it gets all rows.
                # Work around: resize only first row and columns and apply this to all rows/cols
                logger.warning("Performance work around: for tables with more than 1000 cells " +
                               "only the current column is resized and the others use that size.")
                horHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
                self.tableView.resizeColumnToContents(oldCol)
                resizeAllSections(horHeader, self.tableView.columnWidth(oldCol))
        else:
            logger.debug("Setting horizontal resize mode to Interactive and reset header")
            horHeader.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
            resizeAllSections(horHeader, self.config.defaultColWidthCti.configValue)
            horHeader.reset()

        # Restore selection after select
        if selectionWasValid:
            newIndex = self.model.index(oldRow, oldCol)
            if newIndex.isValid():
                logger.debug("Restoring selection: row={}, col={}".format(oldRow, oldCol))
                self.tableView.setCurrentIndex(newIndex)
            else:
                logger.debug("Can't restore selection")



class TableInspectorModel(QtCore.QAbstractTableModel):
    """ Qt table model that gives access to the sliced array,
        To be used in the TableInspector.
    """
    def __init__(self, parent = None):
        """ Constructor

            :param separateFields: If True the fields of a structured array have their own
                                   separate cells.
            :param parent: parent Qt widget.
        """
        super(TableInspectorModel, self).__init__(parent)
        self._nRows = 0
        self._nCols = 0
        self._fieldNames = []
        self._slicedArray = None # can be a masked array or a regular numpy array
        self._rtiInfo = {}

        self._separateFields = True  # User config option
        self._separateFieldOrientation = None # To store which axis is currently separated
        self._numbersInHeader = True
        self._font = None # Default font

        # The following members are simple attributes, they can be changed independently
        self.encoding = 'utf-8'
        self.strFormat = None
        self.numFormat = None
        self.intFormat = None
        self.otherFormat = None
        self.maskFormat = None
        self.textAlignment = None
        self.verAlignment = None


    def updateState(self, slicedArray, rtiInfo, separateFields):
        """ Sets the slicedArray and rtiInfo and other members. This will reset the model.

            Will be called from the tableInspector._drawContents.
        """
        logger.debug("TableInspectorModel.updateState called")
        self.beginResetModel()
        try:
            # The sliced array can be a masked array or a (regular) numpy array.
            # The table works fine with masked arrays, no need to replace the masked values.
            self._slicedArray = slicedArray
            if slicedArray is None:
                self._nRows = 0
                self._nCols = 0
                self._fieldNames = []
            else:
                self._nRows, self._nCols = self._slicedArray.shape
                if self._slicedArray.data.dtype.names:
                    self._fieldNames = self._slicedArray.data.dtype.names
                else:
                    self._fieldNames = []

            self._rtiInfo = rtiInfo
            self._separateFields = separateFields

            # Don't put numbers in the header if the record is of structured type, fields are
            # placed in separate cells and the fake dimension is selected (combo index 0)
            if self._separateFields and self._fieldNames:

                if self._rtiInfo['x-dim'] == FAKE_DIM_NAME:
                    self._separateFieldOrientation = Qt.Horizontal
                    self._numbersInHeader = False
                elif self._rtiInfo['y-dim'] == FAKE_DIM_NAME:
                    self._separateFieldOrientation = Qt.Vertical
                    self._numbersInHeader = False
                else:
                    self._separateFieldOrientation = Qt.Horizontal
                    self._numbersInHeader = True
            else:
                self._separateFieldOrientation = None
                self._numbersInHeader = True

        finally:
            self.endResetModel()


    def _cellValue(self, index):
        """ Returns the data value of the cell at the index (without any string conversion)
        """
        row = index.row()
        col = index.column()
        if (row < 0 or row >= self.rowCount() or col < 0 or col >= self.columnCount()):
            return None

        # The check above should have returned None if the sliced array is None
        assert self._slicedArray is not None, "Sanity check failed."

        nFields = len(self._fieldNames)

        data = self._slicedArray.data
        if self._separateFieldOrientation == Qt.Horizontal:
            dataValue = data[row, col // nFields][self._fieldNames[col % nFields]]
        elif self._separateFieldOrientation == Qt.Vertical:
            dataValue = data[row // nFields, col][self._fieldNames[row % nFields]]
        else:
            dataValue = data[row, col]

        return dataValue


    def _cellMask(self, index):
        """ Returns the data mask of the cell at the index (without any string conversion)
        """
        row = index.row()
        col = index.column()
        if (row < 0 or row >= self.rowCount() or col < 0 or col >= self.columnCount()):
            return None

        # The check above should have returned None if the sliced array is None
        assert self._slicedArray is not None, "Sanity check failed."

        nFields = len(self._fieldNames)
        mask = self._slicedArray.mask
        if is_an_array(mask):
            if self._separateFieldOrientation == Qt.Horizontal:
                maskValue = mask[row, col // nFields][self._fieldNames[col % nFields]]
            elif self._separateFieldOrientation == Qt.Vertical:
                maskValue = mask[row // nFields, col][self._fieldNames[row % nFields]]
            else:
                maskValue = mask[row, col]
        else:
            maskValue = mask

        # Here maskValue can still be a list in case of structured arrays. It can even still be
        # a numpy array in case of a structured array with sub arrays as fields
        if is_an_array(maskValue):
            allMasked = np.all(maskValue)
        else:
            try:
                allMasked = all(maskValue)
            except TypeError as ex:
                allMasked = bool(maskValue)

        return allMasked


    def data(self, index, role = Qt.DisplayRole):
        """ Returns the data at an index for a certain role
        """
        try:
            if role == Qt.DisplayRole:
                return to_string(self._cellValue(index), masked=self._cellMask(index),
                                 decode_bytes=self.encoding, maskFormat=self.maskFormat,
                                 strFormat=self.strFormat, intFormat=self.intFormat,
                                 numFormat=self.numFormat, otherFormat=self.otherFormat)

            elif role == Qt.ToolTipRole:
                fmt = '{!r}' # Use __repr__ for toolbar, except for strings
                return to_string(self._cellValue(index), masked=self._cellMask(index),
                                 decode_bytes=self.encoding, maskFormat=fmt,
                                 strFormat=self.strFormat, intFormat=fmt,
                                 numFormat=fmt, otherFormat=fmt)

            elif role == Qt.FontRole:
                #assert self._font, "Font undefined"
                return self._font

            elif role == Qt.TextColorRole:
                masked = self._cellMask(index)
                if not is_an_array(masked) and masked:
                    return self.missingColor
                else:
                    return self.dataColor

            elif role == Qt.TextAlignmentRole:
                if self.horAlignment == ALIGN_SMART:
                    cellContainsNumber = isinstance(self._cellValue(index), numbers.Number)
                    horAlign = Qt.AlignRight if cellContainsNumber else Qt.AlignLeft
                    return horAlign | self.verAlignment
                else:
                    return self.horAlignment | self.verAlignment

            else:
                return None
        except Exception as ex:
            logger.error("Slot is not exception-safe.")
            logger.exception(ex)
            if DEBUGGING:
                raise


    def flags(self, index):
        """ Returns the item flags for the given index.
        """
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable


    def headerData(self, section, orientation, role):
        """ Returns the header for a section (row or column depending on orientation).
            Reimplemented from QAbstractTableModel to make the headers start at 0.
        """
        if role == Qt.DisplayRole:
            if self._separateFieldOrientation == orientation:

                nFields = len(self._fieldNames)
                varNr = section // nFields
                fieldNr = section % nFields
                header = str(varNr) + ' : ' if self._numbersInHeader else ''
                header += self._fieldNames[fieldNr]
                return header
            else:
                return str(section)
        else:
            return None


    def rowCount(self, parent=None):
        """ The number of rows of the sliced array.
            The 'parent' parameter can be a QModelIndex. It is ignored since the number of
            rows does not depend on the parent.
        """
        if self._separateFieldOrientation == Qt.Vertical:
            return self._nRows * len(self._fieldNames)
        else:
            return self._nRows


    def columnCount(self, parent=None):
        """ The number of columns of the sliced array.
            The 'parent' parameter can be a QModelIndex. It is ignored since the number of
            columns does not depend on the parent.
        """
        if self._separateFieldOrientation == Qt.Horizontal:
            return self._nCols * len(self._fieldNames)
        else:
            return self._nCols


    def getFont(self):
        """ Returns the font that will be returned when data() is called with the Qt.FontRole.
            Can be a QFont or None if no font is set.
        """
        return self._font


    def setFont(self, font):
        """ Sets the font that will be returned when data() is called with the Qt.FontRole.
            Can be a QFont or None if no font is set.
        """
        check_class(font, QtGui.QFont, allow_none=True)
        self._font = font



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Miscellaneous Qt routines.
"""
from __future__ import division, print_function

import logging
import os
import sys
import traceback

from argos import info
from argos.qt.bindings import QtCore, QtWidgets

logger = logging.getLogger(__name__)


################
# QApplication #
################



def initQApplication():
    """ Initializes the QtWidgets.QApplication instance. Creates one if it doesn't exist.

        Sets Argos specific attributes, such as the OrganizationName, so that the application
        persistent settings are read/written to the correct settings file/winreg. It is therefore
        important to call this function at startup. The ArgosApplication constructor does this.

        Returns the application.
    """
    # PyQtGraph recommends raster graphics system for OS-X.
    if 'darwin' in sys.platform:
        graphicsSystem = "raster" # raster, native or opengl
        os.environ.setdefault('QT_GRAPHICSSYSTEM', graphicsSystem)
        logger.debug("Setting QT_GRAPHICSSYSTEM to: {}".format(graphicsSystem))

    app = QtWidgets.QApplication(sys.argv)
    initArgosApplicationSettings(app)
    return app


def initArgosApplicationSettings(app): # TODO: this is Argos specific. Move somewhere else.
    """ Sets Argos specific attributes, such as the OrganizationName, so that the application
        persistent settings are read/written to the correct settings file/winreg. It is therefore
        important to call this function at startup. The ArgosApplication constructor does this.
    """
    assert app, \
        "app undefined. Call QtWidgets.QApplication.instance() or QtCor.QApplication.instance() first."

    logger.debug("Setting Argos QApplication settings.")
    app.setApplicationName(info.REPO_NAME)
    app.setApplicationVersion(info.VERSION)
    app.setOrganizationName(info.ORGANIZATION_NAME)
    app.setOrganizationDomain(info.ORGANIZATION_DOMAIN)


######################
# Exception Handling #
######################

class ResizeDetailsMessageBox(QtWidgets.QMessageBox):
    """ Message box that enlarges when the 'Show Details' button is clicked.
        Can be used to better view stack traces. I could't find how to make a resizeable message
        box but this it the next best thing.

        Taken from:
        http://stackoverflow.com/questions/2655354/how-to-allow-resizing-of-qmessagebox-in-pyqt4
    """
    def __init__(self, detailsBoxWidth=700, detailBoxHeight=300, *args, **kwargs):
        """ Constructor
            :param detailsBoxWidht: The width of the details text box (default=700)
            :param detailBoxHeight: The heights of the details text box (default=700)
        """
        super(ResizeDetailsMessageBox, self).__init__(*args, **kwargs)
        self.detailsBoxWidth = detailsBoxWidth
        self.detailBoxHeight = detailBoxHeight


    def resizeEvent(self, event):
        """ Resizes the details box if present (i.e. when 'Show Details' button was clicked)
        """
        result = super(ResizeDetailsMessageBox, self).resizeEvent(event)

        details_box = self.findChild(QtWidgets.QTextEdit)
        if details_box is not None:
            #details_box.setFixedSize(details_box.sizeHint())
            details_box.setFixedSize(QtCore.QSize(self.detailsBoxWidth, self.detailBoxHeight))

        return result



def handleException(exc_type, exc_value, exc_traceback):

    traceback.format_exception(exc_type, exc_value, exc_traceback)

    logger.critical("Bug: uncaught {}".format(exc_type.__name__),
                    exc_info=(exc_type, exc_value, exc_traceback))
    if info.DEBUGGING:
        logger.info("Quitting application with exit code 1")
        sys.exit(1)
    else:
        # Constructing a QApplication in case this hasn't been done yet.
        if not QtWidgets.qApp:
            _app = QtWidgets.QApplication()

        msgBox = ResizeDetailsMessageBox()
        msgBox.setText("Bug: uncaught {}".format(exc_type.__name__))
        msgBox.setInformativeText(str(exc_value))
        lst = traceback.format_exception(exc_type, exc_value, exc_traceback)
        msgBox.setDetailedText("".join(lst))
        msgBox.setIcon(QtWidgets.QMessageBox.Warning)
        msgBox.exec_()
        logger.info("Quitting application with exit code 1")
        sys.exit(1)



######################
# QSettings routines #
######################


def getWidgetGeom(widget):
    """ Gets the QWindow or QWidget geometry as a QByteArray.

        Since Qt does not provide this directly we hack this by saving it to the QSettings
        in a temporary location and then reading it from the QSettings.

        :param widget: A QWidget that has a saveGeometry() methods
    """
    settings = QtCore.QSettings()
    settings.beginGroup('temp_conversion')
    try:
        settings.setValue("winGeom", widget.saveGeometry())
        return bytes(settings.value("winGeom"))
    finally:
        settings.endGroup()


def getWidgetState(qWindow):
    """ Gets the QWindow or QWidget state as a QByteArray.

        Since Qt does not provide this directly we hack this by saving it to the QSettings
        in a temporary location and then reading it from the QSettings.

        :param widget: A QWidget that has a saveState() methods
    """
    settings = QtCore.QSettings()
    settings.beginGroup('temp_conversion')
    try:
        settings.setValue("winState", qWindow.saveState())
        return bytes(settings.value("winState"))
    finally:
        settings.endGroup()



######################
# Debugging routines #
######################

def printChildren(obj, indent=""):
    """ Recursively prints the children of a QObject. Useful for debugging.
    """
    children=obj.children()
    if children==None:
        return
    for child in children:
        try:
            childName = child.objectName()
        except AttributeError:
            childName = "<no-name>"

        #print ("{}{:10s}: {}".format(indent, childName, child.__class__))
        print ("{}{!r}: {}".format(indent, childName, child.__class__))
        printChildren(child, indent + "    ")


def printAllWidgets(qApplication, ofType=None):
    """ Prints list of all widgets to stdout (for debugging)
    """
    print ("Application's widgets {}".format(('of type: ' + str(ofType)) if ofType else ''))
    for widget in qApplication.allWidgets():
        if ofType is None or isinstance(widget, ofType):
            print ("  {!r}".format(widget))


#####################
# Unsorted routines #
#####################


def widgetSubCheckBoxRect(widget, option):
    """ Returns the rectangle of a check box drawn as a sub element of widget
    """
    opt = QtWidgets.QStyleOption()
    opt.initFrom(widget)
    style = widget.style()
    return style.subElementRect(QtWidgets.QStyle.SE_ViewItemCheckIndicator, opt, widget)



def setWidgetSizePolicy(widget, horPolicy=None, verPolicy=None):
    """ Sets the size policy of a widget.
    """
    sizePolicy = widget.sizePolicy()

    if horPolicy is not None:
        sizePolicy.setHorizontalPolicy(horPolicy)

    if verPolicy is not None:
        sizePolicy.setVerticalPolicy(verPolicy)

    widget.setSizePolicy(sizePolicy)
    return sizePolicy




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.


""" Various definitions, errors and constants that can be used throughout the program

"""
import sys

CONTIGUOUS = 'contiguous'  # contiguous chunking

# String formatting template for numbered dimension names
DIM_TEMPLATE = "dim-{}"
SUB_DIM_TEMPLATE = "subdim-{}"

# Use different unicode character per platform as it looks better.
if sys.platform == 'linux':
    RIGHT_ARROW = "\u2794"
elif sys.platform == 'win32' or sys.platform == 'cygwin':
    RIGHT_ARROW = "\u2794"
elif sys.platform == 'darwin':
    RIGHT_ARROW = "\u279E"
else:
    RIGHT_ARROW = "\u2794"


class InvalidInputError(Exception):
    """ Exception raised when the input is invalid after editing
    """
    pass



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Functionality for Argos

    This is the top level module and should not be imported by sub modules.
    The only way is up in that respect. It imports a few symbols itself for convenience. This
    allows users, for instance, to call argos.browse().
"""

# IMPORTANT: do not do any imports here that (indirectly) import any dependencies (PyQt, numpy, etc)
# The browse function is imported by the argos package, which in turn is imported by setup.py.
# If you import (for instance) numpy here, the setup.py will fail if numpy is not installed.

from .info import VERSION as __version__
from .main import browse




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

r""" Various functions related to logging
"""
import logging
import logging.config
import os.path
import json

from argos.utils.dirs import ensureDirectoryExists, normRealPath, argosLogDirectory
from argos.utils.misc import replaceStringsInDict

logger = logging.getLogger(__name__)


THIS_MODULE_DIR = os.path.dirname(os.path.realpath(__file__))

#
# DEFAULT_LOG_CFG = {
#     'formatters': {
#         'fileFormatter': {
#             'format': '%(asctime)s.%(msecs)03d : pid = %(process)5d : %(threadName)-13s tid = %(thread)5d : %(filename)25s:%(lineno)-4d : %(levelname)-8s: %(message)s',
#             'datefmt': '%Y-%m-%d %H:%M:%S'
#         },
#         'screenFormatter': {
#             'format': '%(asctime)s.%(msecs)03d : pid = %(process)5d : %(threadName)-13s tid = %(thread)5d : %(filename)25s:%(lineno)-4d : %(levelname)-8s: %(message)s',
#             'datefmt': '%Y-%m-%d %H:%M:%S'
#         }
#     },
#
#     'handlers': {
#         'streamHandler': {
#             'class': 'logging.handlers.StreamHandler',
#             'formatter': 'screenFormatter',
#             'stream': 'ext://sys.stderr',
#             'level': 'DEBUG',
#         },
#
#         # The handler that writes to the (operational log file).
#         # This is the log file that is shown when the user clicks the 'show log' file button in the exception dialog.
#         # Therefore do not rename this handler.
#         'currentRunHandler': {
#             'class': 'logging.handlers.FileHandler',
#             'formatter': 'fileFormatter',
#             'filename': '@logDir@/last_run.log',
#             'encoding': 'utf-8',
#             'delay': 'true',  # only create file when used.
#             'mode': 'w',
#             'level': 'DEBUG',
#         },
#
#         'rotatingFileHandler': {  #
#             'class': 'logging.handlers.TimedRotatingFileHandler',
#             'formatter': 'fileFormatter',
#             'filename': '@logDir@/argos.log',
#             'backupCount': 5,
#             'when': 'd',
#             'interval': '1',
#             'encoding': 'utf-8',
#             'delay': 'true',   # only create file when used.
#             'mode': 'a',
#             'level': 'DEBUG',
#         }
#     },
#
#     # This sets level of all log messages, including messages from 3rd party libraries.
#     # In the loggers section below you can override the log levels of (sub)components.
#     'root' : {
#         'level': 'WARNING',
#         'handlers': ['streamHandler', 'currentRunHandler', 'rotatingFileHandler'],
#     },
#
#     'loggers': {
#         'argos': {
#             'level': 'DEBUG',
#         }
#     }
# }

def findStreamHandlersInConfig():
    """ Searches for a handlers with 'stream' their name. Returns a list of handlers.
    """
    rootLogger = logging.getLogger()
    #logger.debug("Searching for Handlers in the root logger haveing 'stream' in their name")
    foundHandlers = []
    for handler in rootLogger.handlers:
        #logger.debug("  handler name: {}".format(handler.name))
        if 'stream' in handler.name.lower():
            foundHandlers.append(handler)

    return foundHandlers



def initLogging(configFileName=None, streamLogLevel=None):
    """ Configures logging given a (JSON) config file name.

        If configFileName is None, the default logging (from iriscc/lib/default_logging.yaml) is
        used.

        :param configFileName: JSON file with log config.
        :param streamLogLevel: If given it overrides the log level of StreamHandlers in the config. All messages below
            this level will be suppressed.
    """
    if configFileName is None:
        configFileName = os.path.join(THIS_MODULE_DIR, "default_logging.json")

    with open(configFileName, 'r') as stream:
        lines = stream.readlines()
        cfgLines = ''.join(lines)

    # Ensure the directory exists if @logDir@ is in the JSON file.
    logDir = argosLogDirectory()
    if '@logDir@' in cfgLines:
        ensureDirectoryExists(logDir)

    configDict = json.loads(cfgLines)
    configDict = replaceStringsInDict(configDict, "@logDir@", logDir)

    logging.config.dictConfig(configDict)

    if streamLogLevel:
        # Using getLevelName to get the level number. This undocumented behavior has been upgraded
        # to documented behavior in Python 3.4.2.
        # See https://docs.python.org/3.4/library/logging.html#logging.getLevelName
        levelNr = logging.getLevelName(streamLogLevel.upper()) - 1
        #logging.disable(levelNr)

        for streamHandler in findStreamHandlersInConfig():
            logger.debug("Setting log level to {} in handler: {} ".format(levelNr, streamHandler))
            streamHandler.setLevel(levelNr)

    logging.info("Initialized logging from: '{}'".format(normRealPath(configFileName)))
    logging.info("Default location of log files: '{}'".format(logDir))


def log_dictionary(dictionary, msg='', logger=None, level='debug', item_prefix='    '):
    """ Writes a log message with key and value for each item in the dictionary.

        :param dictionary: the dictionary to be logged
        :type dictionary: dict
        :param name: An optional message that is logged before the contents
        :type name: string
        :param logger: A logging.Logger object to log to. If not set, the 'main' logger is used.
        :type logger: logging.Logger or a string
        :param level: log level. String or int as described in the logging module documentation.
            Default: 'debug'.
        :type level: string or int
        :param item_prefix: String that will be prefixed to each line. Default: two spaces.
        :type item_prefix: string
    """
    level_nr = logging.getLevelName(level.upper())

    if logger is None:
        logger = logging.getLogger('main')

    if msg :
        logger.log(level_nr, "Logging dictionary: {}".format(msg))

    if not dictionary:
        logger.log(level_nr,"{}<empty dictionary>".format(item_prefix))
        return

    max_key_len = max([len(k) for k in dictionary.keys()])

    for key, value in sorted(dictionary.items()):
        logger.log(level_nr, "{0}{1:<{2}s} = {3}".format(item_prefix, key, max_key_len, value))


def make_log_format(
        ascTime = True,
        processId = False,
        threadName = False,
        threadId = False,
        fileLine = True,
        loggerName = False,
        level = True):
    """ Creates a format string to use in logging.basicConfig.

        Use example:
            logging.basicConfig(level="DEBUG", format=makeLogFormat(fileLine=False))
    """
    parts = []
    if ascTime:
        parts.append('%(asctime)s')

    if processId:
        parts.append('pid=%(process)5d')

    if threadName:
        parts.append('%(threadName)15s')

    if threadId:
        parts.append('id=0x%(thread)x')

    if fileLine:
        parts.append('%(filename)25s:%(lineno)-4d')

    if loggerName:
        parts.append("%(name)30s")

    if level:
        parts.append('%(levelname)-7s')

    parts.append('%(message)s')
    return " : ".join(parts)



# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Classes for registering plugins.
"""
import enum
import logging, inspect, os, sys
import traceback

from argos.info import DEBUGGING
from argos.utils.cls import import_symbol, check_is_a_string, type_name, check_class
from argos.reg.tabmodel import BaseItem, BaseItemStore, BaseTableModel
from argos.qt import Qt, QtGui
from argos.utils.misc import string_to_identifier

logger = logging.getLogger(__name__)


QCOLOR_REGULAR = QtGui.QColor('black')
QCOLOR_NOT_IMPORTED = QtGui.QColor('grey')
QCOLOR_ERROR = QtGui.QColor('red')

# Used to determine the type of editor needed
@enum.unique
class RegType(enum.Enum):
    String = 0
    ColorStr = 1
    ShortCut = 2


def nameToIdentifier(fullName):
    """ Constructs the regItem identifier given its full name
    """
    return string_to_identifier(fullName, white_space_becomes='')


class BaseRegItem(BaseItem):
    """ Represents a class that is registered in the registry.

        Each registry item (RegItem) can import its class. If the import fails, the exception info
        is put into the exception property. The underlying class is not imported by default;
        use tryImportClass or getClass() for this.

        Some of the underlying class's attributes, such as the docstring, are made available as
        properties of the RegItem as well. If the class is not yet imported, they return None.
    """
    FIELDS = ['name', 'absClassName', 'pythonPath']
    TYPES = [RegType.String, RegType.String, RegType.String]
    LABELS = ['Name', 'Class', 'Python path']
    STRETCH = [False, False, True]

    def __init__(self, name='', absClassName='', pythonPath=''):
        """ Constructor.

            :param name: fullName comprising of library and name, separated by a slash.
                Can contain spaces. E.g.: 'library name/My Widget'
                Must be unique when spaces are removed and converted to lower case.
            :param absClassName: absolute name of the underlying class. Must include the full
                path of packages and module. E.g.: 'argos.plugins.rti.ncdf.NcdfFileInspector'
            :param pythonPath: directory that will be added to the sys.path before importing.
                Can be multiple directories separated by a colon (:)
        """
        super(BaseRegItem, self).__init__()

        self._data = {'name': name, 'absClassName': absClassName, 'pythonPath': pythonPath}

        self._cls = None # The underlying class. Not yet imported.
        self._triedImport = False
        self._exception = None # Any exception that occurs during the class import


    def __repr__(self):
        return "<{} (Ox{:x}): {!r}>".format(type_name(self), id(self), self.name)


    @property
    def identifier(self):
        """ Identifier. Should be unique.

            Is the name with white space removed.
        """
        return nameToIdentifier(self._data['name'])


    @property
    def name(self):
        """ Name of the registered plug in.
        """
        return self._data['name']


    @property
    def absClassName(self):
        """ Absolute name of the underlying class.

        Must include the full path of packages and module.
        E.g.: 'argos.plugins.rti.ncdf.NcdfFileInspector'
        """
        return self._data['absClassName']


    @property
    def pythonPath(self):
        """ Directory that will be added to the sys.path before importing.
            Can be multiple directories separated by a colon (:)
        """
        return self._data['pythonPath']


    @property
    def library(self):
        """ The fullName minus the last part (the name).
            Used to group libraries together, for instance in menus.
        """
        return os.path.dirname(self.absClassName)


    def splitName(self):
        """ Returns (self.library, self.name) tuple but is more efficient than calling both
            properties separately.
        """
        return os.path.split(self.absClassName)


    @property
    def cls(self):
        """ Returns the underlying class.
            Returns None if the class was not imported or import failed.
        """
        return self._cls


    @property
    def docString(self):
        """ A cleaned up version of the doc string of the registered class.
            Can serve as backup in case descriptionHtml is empty.
        """
        return inspect.cleandoc('' if self.cls is None else self.cls.__doc__)


    @property
    def descriptionHtml(self):
        """ HTML help describing the class. For use in the detail editor.
        """
        if self.cls is None:
            return None
        elif hasattr(self.cls, 'descriptionHtml'):
            return self.cls.descriptionHtml()
        else:
            return ''


    @property
    def triedImport(self):
        """ Returns True if the class has been imported (either successfully or not)
        """
        return self._triedImport

    @triedImport.setter
    def triedImport(self, value):
        """ Set to true if the class has been imported (either successfully or not)
        """
        self._triedImport = value


    @property
    def successfullyImported(self):
        """ Returns True if the import was a success, False if an exception was raised.
            Returns None if the class was not yet imported.
        """
        if self.triedImport:
            return self.exception is None
        else:
            return None

    @property
    def exception(self):
        """ The exception that occurred during the class import.
            Returns None if the import was successful.
        """
        return self._exception


    def tryImportClass(self):
        """ Tries to import the registered class.
            Will set the exception property if and error occurred.
        """
        logger.debug("Importing: {}".format(self.absClassName))
        self._triedImport = True
        self._exception = None
        self._cls = None
        try:
            for pyPath in self.pythonPath.split(';'):
                if pyPath and pyPath not in sys.path:
                    logger.debug("Appending {!r} to the PythonPath.".format(pyPath))
                    sys.path.append(pyPath)
            self._cls = import_symbol(self.absClassName)
        except Exception as ex:
            # Add traceback in human readable form to the exception for later printing.
            # This is a hack for Python-2 to store the traceback later use. See https://stackoverflow.com/a/11415140/
            ex.traceBackString = traceback.format_exc()
            self._exception = ex
            logger.warning("Unable to import {!r}: {}".format(self.absClassName, ex))
            logger.debug("Traceback: {}".format(ex.traceBackString))


    def getClass(self, tryImport=True):
        """ Gets the underlying class. Tries to import if tryImport is True (the default).
            Returns None if the import has failed (the exception property will contain the reason)
        """
        if not self.triedImport and tryImport:
            self.tryImportClass()

        return self._cls



class BaseRegistry(BaseItemStore):
    """ Class that maintains the collection of registered classes (plugins).

        It can load or store its classes in the persistent settings. It can also create a default
        set of plugins that can be used initially, the first time the program is executed.

        The BaseRegistry can only store items of one type (ClassRegItem). Descendants will
        store their own type. For instance the InspectorRegistry will store InspectorRegItem
        items. This makes serialization easier.
    """
    ITEM_CLASS = BaseRegItem

    @property
    def registryName(self):
        """ # Human readable name for this registry. Please override.
        """
        raise NotImplementedError()


    def getItemById(self, identifier):
        """ Gets a registered item given its identifier. Returns None if not found.
        """
        #logger.debug("Searching for '{}' in items".format(identifier))
        for item in self._items:
            #logger.debug("    Item: {}".format(item.identifier))
            if item.identifier == identifier:
                return item

        return None


    def getDefaultItems(self):
        """ Returns a list with the default plugins in the registry.
            This is used initialize the application plugins when there are no saved settings,
            for instance the first time the application is started.
            The base implementation returns an empty list but other registries should override it.
        """
        raise NotImplementedError


    def createTableModel(self, parent=None):
        """ Creates a BaseRegistryModel that has self as an item store.

            Descendants can override so they can create specialized types of registry models
        """
        return BaseRegistryModel(store=self, parent=parent)



class BaseRegistryModel(BaseTableModel):
    """ Table model that holds a BaseRegistry store
    """

    def __init__(self, store, parent=None):
        """ Constructor.

            :param store: Underlying data store, must descent from BaseRegistry
            :param parent: Parent widget
        """
        super(BaseRegistryModel, self).__init__(store, parent)
        check_class(store, BaseRegistry)

        self.regularBrush = QtGui.QBrush(QCOLOR_REGULAR)
        self.notImportedBrush = QtGui.QBrush(QCOLOR_NOT_IMPORTED)
        self.errorBrush = QtGui.QBrush(QCOLOR_ERROR)


    def data(self, index, role=Qt.DisplayRole):
        """ Returns the data stored under the given role for the item referred to by the index.
        """
        if not index.isValid():
            return None

        if role == Qt.ForegroundRole:
            item = self._store.items[index.row()]

            if item.successfullyImported is None:
                return self.notImportedBrush
            elif item.successfullyImported:
                return self.regularBrush
            else:
                return self.errorBrush
        else:
            return super(BaseRegistryModel, self).data(index, role=role)


    def setData(self, index, value, role=Qt.EditRole):
        """ Sets the role data for the item at index to value.
        """
        result = super(BaseRegistryModel, self).setData(index, value, role=role)

        if not index.isValid():
            return False

        if role != Qt.EditRole:
            return False

        regItem = self._store.items[index.row()]
        regItem.tryImportClass()

        self.emitDataChanged(regItem)
        return result


    def tryImportRegItem(self, regItem):
        """ Tries to import a registry item (plugin)
        """
        check_class(regItem, self.store.ITEM_CLASS)
        logger.debug("Importing {}...".format(regItem.name))
        regItem.tryImportClass()
        self.emitDataChanged(regItem)



# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

"""
    Main window functionality

"""


import base64
import cProfile
import logging
import os.path
import pstats
from functools import partial

import numpy as np

from argos.collect.collector import Collector
from argos.config.abstractcti import AbstractCti
from argos.config.configtreemodel import ConfigTreeModel
from argos.config.configtreeview import ConfigWidget
from argos.info import DEBUGGING, PROJECT_NAME, PROFILING, EXIT_CODE_RESTART
from argos.inspector.abstract import AbstractInspector, UpdateReason
from argos.inspector.errormsg import ErrorMsgInspector
from argos.inspector.selectionpane import InspectorSelectionPane
from argos.qt import Qt, QUrl, QtCore, QtGui, QtWidgets, QtSignal, QtSlot
from argos.qt.misc import getWidgetGeom, getWidgetState
from argos.reg.basereg import nameToIdentifier
from argos.reg.dialog import PluginsDialog
from argos.repo.iconfactory import RtiIconFactory
from argos.repo.registry import RtiRegistry
from argos.repo.repotreeview import RepoWidget
from argos.repo.testdata import createArgosTestData
from argos.utils.cls import check_class, check_is_a_sequence
from argos.utils.dirs import argosConfigDirectory, argosLogDirectory
from argos.utils.misc import string_to_identifier
from argos.utils.moduleinfo import versionStrToTuple
from argos.widgets.aboutdialog import AboutDialog
from argos.widgets.testwalkdialog import TestWalkDialog

logger = logging.getLogger(__name__)

# The main window inherits from a Qt class, therefore it has many
# ancestors public methods and attributes.
# pylint: disable=R0901, R0902, R0904, W0201


class MainWindow(QtWidgets.QMainWindow):
    """ Main application window.
    """
    __numInstances = 0

    # Emitted when the inspector has changed.
    sigInspectorChanged = QtSignal(object)  # InspectorRegItem or None
    sigShowMessage = QtSignal(str) # Message to the user in the inspector-selection panel

    def __init__(self, argosApplication):
        """ Constructor
            :param reset: If true the persistent settings, such as column widths, are reset.
        """
        super(MainWindow, self).__init__()
        self._windowNumber = MainWindow.__numInstances # Used only for debugging
        MainWindow.__numInstances += 1

        if PROFILING:
            # Profiler that measures the drawing of the inspectors.
            self._profFileName = "inspectors.prof"

            logger.warning("Profiling is on for {}. This may cost a bit of CPU time.")
            self._profiler = cProfile.Profile()

        self.testWalkDialog = TestWalkDialog(mainWindow=self, parent=self)  # don't show yet

        self._collector = None
        self._inspector = ErrorMsgInspector(
            self._collector, "No inspector yet. Please select one from the menu.")
        self._inspector.sigShowMessage.connect(self.sigShowMessage)
        self._inspector.sigUpdated.connect(self.testWalkDialog.setTestResult)
        self._inspectorRegItem = None # The registered inspector item a InspectorRegItem)

        self._argosApplication = argosApplication
        self._configTreeModel = ConfigTreeModel()
        self._inspectorStates = {}  # keeps track of earlier inspector states
        self._currentTestName = None
        self._failedTests = []

        self.setDockNestingEnabled(False)
        self.setCorner(Qt.TopLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.TopRightCorner, Qt.TopDockWidgetArea)
        self.setCorner(Qt.BottomRightCorner, Qt.BottomDockWidgetArea)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setUnifiedTitleAndToolBarOnMac(True)
        #self.setDocumentMode(True) # Look of tabs as Safari on OS-X (disabled, ugly)
        self.resize(1300, 700)  # Assumes minimal resolution of 1366 x 768

        # Move window to the center of the screen
        desktopRect = QtWidgets.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()
        self.move(round(center.x() - self.width () * 0.5), round(center.y() - self.height() * 0.5))

        self.__setupActions()
        self.__setupMenus()
        self.__setupViews()
        self.__setupDockWidgets()

        for detailsPane in self.repoWidget.detailDockPanes:
            detailsPane.sigUpdated.connect(self.testWalkDialog.setTestResult)


    def finalize(self):
        """ Is called before destruction (when closing).
            Can be used to clean-up resources.
        """
        logger.debug("Finalizing: {}".format(self))

        self.testWalkDialog.finalize()

        # Disconnect signals
        self.collector.sigContentsChanged.disconnect(self.collectorContentsChanged)
        self._configTreeModel.sigItemChanged.disconnect(self.configContentsChanged)
        self.sigInspectorChanged.disconnect(self.inspectorSelectionPane.updateFromInspectorRegItem)
        self.sigShowMessage.disconnect(self.inspectorSelectionPane.showMessage)
        self._collector.sigShowMessage.disconnect(self.sigShowMessage)
        self._inspector.sigShowMessage.disconnect(self.sigShowMessage)
        self._inspector.sigUpdated.disconnect(self.testWalkDialog.setTestResult)

        for detailsPane in self.repoWidget.detailDockPanes:
            detailsPane.sigUpdated.disconnect(self.testWalkDialog.setTestResult)

        if PROFILING:
            logger.info("Saving profiling information to {}"
                        .format(os.path.abspath(self._profFileName)))
            profStats = pstats.Stats(self._profiler)
            profStats.dump_stats(self._profFileName)

        self.inspector.finalize()


    def __setupViews(self):
        """ Creates the UI widgets.
        """
        self._collector = Collector(self.windowNumber)
        self._collector.sigShowMessage.connect(self.sigShowMessage)

        self.configWidget = ConfigWidget(self._configTreeModel)
        self.repoWidget = RepoWidget(self.argosApplication.repo, self.collector)

        # self._configTreeModel.insertItem(self.repoWidget.repoTreeView.config) # No configurable items yet

        # Define a central widget that will be the parent of the inspector widget.
        # We don't set the inspector directly as the central widget to retain the size when the
        # inspector is changed.
        self.mainWidget = QtWidgets.QWidget()
        self.mainLayout = QtWidgets.QVBoxLayout(self.mainWidget)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)
        self.setCentralWidget(self.mainWidget)

        self.topPane = QtWidgets.QFrame()
        # self.topPane.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Raised)
        # self.topPane.setLineWidth(1)
        self.mainLayout.addWidget(self.topPane)

        self.topLayout = QtWidgets.QHBoxLayout(self.topPane)
        self.topLayout.setContentsMargins(0, 0, 0, 0)
        self.topLayout.setSpacing(0)

        self.inspectorSelectionPane = InspectorSelectionPane(self.inspectorActionGroup)
        self.topLayout.addWidget(self.inspectorSelectionPane)
        self.sigInspectorChanged.connect(self.inspectorSelectionPane.updateFromInspectorRegItem)
        self.sigShowMessage.connect(self.inspectorSelectionPane.showMessage)

        showInspectorMenuAction = QtWidgets.QAction("ShowInspectorMenu", self,
            triggered=self.inspectorSelectionPane.menuButton.showMenu, checkable=False)
        showInspectorMenuAction.setShortcut("Ctrl+I")
        self.addAction(showInspectorMenuAction)

        self.wrapperWidget = QtWidgets.QWidget()
        self.mainLayout.addWidget(self.wrapperWidget)

        self.wrapperLayout = QtWidgets.QVBoxLayout(self.wrapperWidget)
        self.wrapperLayout.setContentsMargins(0, 0, 0, 0)
        self.wrapperLayout.setSpacing(0)
        self.wrapperLayout.addWidget(self.inspector)

        # Must be after setInspector since that already draws the inspector
        self.collector.sigContentsChanged.connect(self.collectorContentsChanged)
        self._configTreeModel.sigItemChanged.connect(self.configContentsChanged)

        # Populate table headers menu
        self.__addTableHeadersSubMenu("Data", self.repoWidget.repoTreeView)
        self.__addTableHeadersSubMenu("Settings", self.configWidget.configTreeView)
        self.__addTableHeadersSubMenu(
            "Properties", self.repoWidget.propertiesPane.table)
        self.__addTableHeadersSubMenu(
            "Attributes", self.repoWidget.attributesPane.table)


    def __addTableHeadersSubMenu(self, menuTitle, treeView):
        """ Adds a sub menu to the View | Table Headers menu with actions to show/hide columns
        """
        subMenu = self.tableHeadersMenu.addMenu(menuTitle)
        for action in treeView.getHeaderContextMenuActions():
            subMenu.addAction(action)


    def __setupActions(self):
        """ Creates actions that are always usable, even if they are not added to the main menu.

            Some actions are only added to the menu in debugging mode.
        """
        self.showTestWalkDialogAction = QtWidgets.QAction("Test Walk...", self)
        self.showTestWalkDialogAction.setToolTip("Shows test-walk dialog windows.")
        self.showTestWalkDialogAction.setShortcut("Ctrl+T")
        self.showTestWalkDialogAction.triggered.connect(self.showTestWalkDialog)
        self.addAction(self.showTestWalkDialogAction)

        # The action added to the menu in the repopulateWindowMenu method, which is called by
        # the ArgosApplication object every time a window is added or removed.
        self.activateWindowAction = QtWidgets.QAction("Window #{}".format(self.windowNumber), self)
        self.activateWindowAction.triggered.connect(self.activateAndRaise)
        self.activateWindowAction.setCheckable(True)
        #self.addAction(self.activateWindowAction)

        if self.windowNumber <= 9:
            self.activateWindowAction.setShortcut(QtGui.QKeySequence(
                "Alt+{}".format(self.windowNumber)))

        self.addTestDataAction = QtWidgets.QAction("Add Test Data", self)
        self.addTestDataAction.setToolTip("Add in-memory test data to the tree.")
        self.addTestDataAction.setShortcut("Meta+A")
        self.addTestDataAction.triggered.connect(self.addTestData)
        self.addAction(self.addTestDataAction)

        self.myTestAction = QtWidgets.QAction("My Test", self)
        self.myTestAction.setToolTip("Ad-hoc test procedure for debugging.")
        self.myTestAction.setShortcut("Meta+T")
        self.myTestAction.triggered.connect(self.myTest)
        self.addAction(self.myTestAction)


    def __setupMenus(self):
        """ Sets up the main menu.
        """
        if True:
            # Don't use self.menuBar(), on OS-X this is not shared across windows.
            # See: http://qt-project.org/doc/qt-4.8/qmenubar.html#details
            # And:http://qt-project.org/doc/qt-4.8/qmainwindow.html#menuBar
            menuBar = QtWidgets.QMenuBar() # Make a menu without parent.
            self.setMenuBar(menuBar)
        else:
            menuBar = self.menuBar()

        ### File Menu ###

        fileMenu = menuBar.addMenu("&File")

        fileMenu.addAction("&New Window", self.cloneWindow, QtGui.QKeySequence("Ctrl+N"))
        fileMenu.addAction("Close &Window", self.close, QtGui.QKeySequence.Close)

        fileMenu.addSeparator()

        action = fileMenu.addAction("Browse Directory...",
            lambda: self.openFiles(fileMode = QtWidgets.QFileDialog.Directory))
        action.setShortcut(QtGui.QKeySequence("Ctrl+B"))

        action = fileMenu.addAction("&Open Files...",
            lambda: self.openFiles(fileMode = QtWidgets.QFileDialog.ExistingFiles))
        action.setShortcut(QtGui.QKeySequence("Ctrl+O"))

        self.openAsMenu = fileMenu.addMenu("Open As")
        self.openAsMenu.aboutToShow.connect(self._repopulateOpenAsMenu)

        self.openRecentMenu = fileMenu.addMenu("Open Recent")
        self.openRecentMenu.aboutToShow.connect(self._repopulateOpenRecentMenu)

        fileMenu.addSeparator()

        fileMenu.addSeparator()
        fileMenu.addAction("E&xit", self.argosApplication.quit, 'Ctrl+Q')

        ### View Menu ###

        self.viewMenu = menuBar.addMenu("&View")

        self.inspectorActionGroup = self.__createInspectorActionGroup(self)
        for action in self.inspectorActionGroup.actions():
            self.viewMenu.addAction(action)

        self.viewMenu.addSeparator()
        self.panelsMenu = self.viewMenu.addMenu("&Panels")
        self.tableHeadersMenu = self.viewMenu.addMenu("&Table Headers")

        self.viewMenu.addSeparator()
        self.viewMenu.addAction(self.showTestWalkDialogAction)

        ### Config Menu ###

        self.configMenu = menuBar.addMenu("Configure")

        app = self.argosApplication
        action = self.configMenu.addAction("&File Format Plugins...",
            lambda: self.execPluginsDialog("File Formats", app.rtiRegistry))
        action.setShortcut(QtGui.QKeySequence("Ctrl+P"))

        action = self.configMenu.addAction("&Inspector Plugins...",
            lambda: self.execPluginsDialog("Inspectors", app.inspectorRegistry))

        self.configMenu.addSeparator()

        self.configMenu.addAction(
            "Show Config Files...",
            lambda: self.openInExternalApp(argosConfigDirectory()))

        ### Window Menu ###

        # Will be populated in repopulateWindowMenu
        self.windowMenu = menuBar.addMenu("&Window")

        ### Help Menu ###
        menuBar.addSeparator()
        helpMenu = menuBar.addMenu("&Help")
        helpMenu.addAction(
            "&Online Documentation...",
            lambda: self.openInWebBrowser("https://github.com/titusjan/argos#argos"))

        helpMenu.addSeparator()

        helpMenu.addAction(
            "Show Log Files...",
            lambda: self.openInExternalApp(argosLogDirectory()))

        helpMenu.addAction('&About...', self.about)

        if DEBUGGING:

            helpMenu.addSeparator()
            helpMenu.addAction(self.addTestDataAction)

            helpMenu.addAction(
                "Quick Walk &Current Node",
                lambda: self.testWalkDialog.walkCurrentRepoNode(False, False), "Meta+Q")
            helpMenu.addAction(
                "Walk &All Nodes",
                lambda: self.testWalkDialog.walkAllRepoNodes(True, True), "Meta+W")  # meta works on MacOs

            helpMenu.addSeparator()
            helpMenu.addAction(self.myTestAction)



    def __createInspectorActionGroup(self, parent):
        """ Creates an action group with 'set inspector' actions for all installed inspector.
        """
        actionGroup = QtWidgets.QActionGroup(parent)
        actionGroup.setExclusive(True)

        for item in self.argosApplication.inspectorRegistry.items:
            logger.debug("__createInspectorActionGroup item: {} {}".format(item.identifier, item._data))
            setAndDrawFn = partial(self.setAndDrawInspectorById, item.identifier)
            action = QtWidgets.QAction(item.name, self, triggered=setAndDrawFn, checkable=True)
            action.setData(item.identifier)
            if item.shortCut:
                try:
                    keySeq = QtGui.QKeySequence(item.shortCut.strip())
                except Exception as ex:
                    logger.warning("Unable to create shortcut from: '{}".format(item.shortCut))
                else:
                    action.setShortcut(QtGui.QKeySequence(keySeq))

            actionGroup.addAction(action)

        return actionGroup


    def __setupDockWidgets(self):
        """ Sets up the dock widgets. Must be called after the menu is setup.
        """
        #self.dockWidget(self.currentInspectorPane, "Current Inspector", Qt.LeftDockWidgetArea)

        self.dockWidget(self.repoWidget, "Data", Qt.LeftDockWidgetArea)
        self.dockWidget(self.collector, "Collector", Qt.BottomDockWidgetArea)
        # TODO: if the title == "Settings" it won't be added to the view menu (2020-03-29 On OS-X it seems to work now)
        self.dockWidget(self.configWidget, "Settings", Qt.RightDockWidgetArea)



    ##############
    # Properties #
    ##############

    @property
    def windowNumber(self):
        """ The instance number of this window.
        """
        return self._windowNumber

    @property
    def inspectorRegItem(self):
        """ The InspectorRegItem that has been selected. Contains an InspectorRegItem.
            Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem

    @property
    def inspectorId(self):
        """ The ID of the inspector registry item that has been selected.
            E.g. ''. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.identifier if self._inspectorRegItem else None

    @property
    def inspectorName(self):
        """ The name of the inspector registry item that has been selected.
            E.g. 'Table'. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.name if self._inspectorRegItem else None

    @property
    def inspectorFullName(self):
        """ The full name of the inspector registry item that has been selected.
            E.g. 'Qt/Table'. Can be None (e.g. at start-up).
        """
        return self._inspectorRegItem.name if self._inspectorRegItem else None

    @property
    def inspector(self):
        """ The inspector widget of this window. Can be None (e.g. at start-up).
        """
        return self._inspector

    @property
    def collector(self):
        """ The collector widget of this window
        """
        return self._collector

    @property
    def argosApplication(self):
        """ The ArgosApplication to which this window belongs.
        """
        return self._argosApplication

    ###########
    # Methods #
    ###########


    def _repopulateOpenRecentMenu(self, *args, **kwargs):
        """ Clear the window menu and fills it with the actions of the actionGroup
        """
        logger.debug("Called _repopulateOpenRecentMenu")
        rtiIconFactory = RtiIconFactory.singleton()

        for action in self.openRecentMenu.actions():
            self.openRecentMenu.removeAction(action)

        # Count duplicate basename. These will be added with their full path.
        baseNameCount = {}
        for _timeStamp, fileName, _rtiRegName in self._argosApplication.getRecentFiles():
            _, baseName = os.path.split(fileName)
            if baseName in baseNameCount:
                baseNameCount[baseName] += 1
            else:
                baseNameCount[baseName] = 1

        # List returned by getRecentFiles is already sorted.
        for _timeStamp, fileName, rtiRegItemName in self._argosApplication.getRecentFiles():

            regItemId = nameToIdentifier(rtiRegItemName)
            rtiRegItem = self.argosApplication.rtiRegistry.getItemById(regItemId)

            # Bit of a hack to use a Directory RegItem, which is not in the registry.
            if rtiRegItem is None:
                directoryRtiRegItem = self.argosApplication.rtiRegistry.DIRECTORY_REG_ITEM
                if rtiRegItemName == 'Directory':
                    rtiRegItem = directoryRtiRegItem

            if rtiRegItem and not rtiRegItem.triedImport:
                rtiRegItem.tryImportClass()

            def createTrigger():
                "Function to create a closure with the regItem"
                _fileNames = [fileName] # keep reference in closure
                _rtiRegItem = rtiRegItem # keep reference in closure
                return lambda: self.openFiles(_fileNames, rtiRegItem=_rtiRegItem)

            dirName, baseName = os.path.split(fileName)
            fileLabel = fileName if baseNameCount[baseName] > 1 else baseName

            action = QtWidgets.QAction(fileLabel, self, enabled=True, triggered=createTrigger())
            action.setToolTip(fileName)
            if rtiRegItem is not None:
                action.setIcon(rtiRegItem.decoration)
            else:
                # Reserve space
                action.setIcon(rtiIconFactory.getIcon(RtiIconFactory.TRANSPARENT, False))

            self.openRecentMenu.addAction(action)


    def _repopulateOpenAsMenu(self, *args, **kwargs):
        """ Clear the window menu and fills it with the actions of the actionGroup
        """
        logger.debug("Called repopulateOpenAsMenu")

        for action in self.openAsMenu.actions():
            self.openAsMenu.removeAction(action)

        rtiRegistry = self.argosApplication.rtiRegistry
        for rtiRegItem in (rtiRegistry.items + rtiRegistry.extraItemsForOpenAsMenu()):
            if not rtiRegItem.triedImport:
                rtiRegItem.tryImportClass()

            def createTrigger():
                "Function to create a closure with the regItem"
                _rtiRegItem = rtiRegItem # keep reference in closure
                return lambda: self.openFiles(rtiRegItem=_rtiRegItem,
                                              fileMode = QtWidgets.QFileDialog.ExistingFiles,
                                              caption="Open {}".format(_rtiRegItem.name))

            action = QtWidgets.QAction("{}...".format(rtiRegItem.name), self,
                enabled=bool(rtiRegItem.successfullyImported),
                triggered=createTrigger(), icon=rtiRegItem.decoration)

            self.openAsMenu.addAction(action)


    def repopulateWindowMenu(self, actionGroup):
        """ Clear the window menu and fills it with the actions of the actionGroup
        """
        for action in self.windowMenu.actions():
            self.windowMenu.removeAction(action)

        for action in actionGroup.actions():
            self.windowMenu.addAction(action)


    def dockWidget(self, widget, title, area):
        """ Adds a widget as a docked widget.
            Returns the added dockWidget
        """
        assert widget.parent() is None, "Widget already has a parent"

        dockWidget = QtWidgets.QDockWidget(title, parent=self)
        # Use dock2 as name to reset at upgrade
        dockWidget.setObjectName("dock2_" + string_to_identifier(title)) # Use doc
        dockWidget.setWidget(widget)
        self.addDockWidget(area, dockWidget)

        self.panelsMenu.addAction(dockWidget.toggleViewAction())
        return dockWidget


    def updateWindowTitle(self):
        """ Updates the window title frm the window number, inspector, etc
            Also updates the Window Menu
        """
        title = "{} #{} | {}".format(self.inspectorName, self.windowNumber, PROJECT_NAME)

        # Display settings file name in title bar if it's not the default
        settingsFile = os.path.basename(self.argosApplication.settingsFile)
        if settingsFile != 'settings.json':
            title = "{} -- {}".format(title, settingsFile)

        self.setWindowTitle(title)
        #self.activateWindowAction.setText("{} window".format(self.inspectorName, self.windowNumber))
        self.activateWindowAction.setText("{} window".format(self.inspectorName))


    def getInspectorActionById(self, identifier):
        """ Sets the inspector and draw the contents
            Triggers the corresponding action so that it is checked in the menus.
        """
        for action in self.inspectorActionGroup.actions():
            if action.data() == identifier:
                return action
        raise KeyError("No action found with ID: {!r}".format(identifier))


    def setAndDrawInspectorById(self, identifier):
        """ Sets the inspector and draw the contents.

            Does NOT trigger any actions, so the check marks in the menus are not updated. To
            achieve this, the user must update the actions by hand (or call
            getInspectorActionById(identifier).trigger() instead).
        """
        self.setInspectorById(identifier)

        # Show dialog box if import was unsuccessful.
        regItem = self.inspectorRegItem
        if regItem and not regItem.successfullyImported:
            msg = "Unable to import {} inspector.\n{}".format(regItem.identifier, regItem.exception)
            QtWidgets.QMessageBox.warning(self, "Warning", msg)
            logger.warning(msg)

        self.drawInspectorContents(reason=UpdateReason.INSPECTOR_CHANGED)


    def setInspectorById(self, identifier):
        """ Sets the central inspector widget given a inspector ID.

            If identifier is None, the inspector will be unset. Otherwise it will lookup the
            inspector class in the registry. It will raise a KeyError if the ID is not found there.

            It will do an import of the inspector code if it's loaded for the first time. If the
            the inspector class cannot be imported a warning is logged and the inspector is unset.

            NOTE: does not draw the new inspector, this is the responsibility of the caller.
            Also, the corresponding action is not triggered.

            Emits the sigInspectorChanged(self.inspectorRegItem)
        """
        logger.debug("Setting inspector: {}".format(identifier))

        # Use the identifier to find a registered inspector and set self.inspectorRegItem.
        # Then create an inspector object from it.

        oldInspectorRegItem = self.inspectorRegItem
        oldInspector = self.inspector

        inspector = None
        errMsg = None

        if not identifier:
            errMsg = "No inspector selected. Please select one from menu."
            self._inspectorRegItem = None
        else:
            inspectorRegistry = self.argosApplication.inspectorRegistry
            inspectorRegItem = inspectorRegistry.getItemById(identifier)  #

            self._inspectorRegItem = inspectorRegItem
            if inspectorRegItem is None:
                errMsg = "No {!r} inspector found. Please select one from menu.".format(identifier)
            else:
                try:
                    inspector = inspectorRegItem.create(self.collector, tryImport=True)
                except ImportError as ex:
                    # Only log the error. No dialog box or user interaction here because this
                    # function may be called at startup.
                    errMsg = "Unable to create {!r} inspector because {}"\
                        .format(inspectorRegItem.identifier, ex)
                    logger.warning(errMsg, exc_info=DEBUGGING)
                    self.getInspectorActionById(identifier).setEnabled(False)

        ###### Set self.inspector ######

        check_class(inspector, AbstractInspector, allow_none=True)

        logger.debug("Disabling updates.")
        self.setUpdatesEnabled(False)
        try:
            # Delete old inspector
            self._storeInspectorState(oldInspectorRegItem, oldInspector)

            oldInspector.sigShowMessage.disconnect(self.sigShowMessage)
            oldInspector.sigUpdated.disconnect(self.testWalkDialog.setTestResult)
            oldInspector.finalize()
            self.wrapperLayout.removeWidget(oldInspector)
            oldInspector.deleteLater()

            # Set new inspector, update collector widgets and the config tree
            oldBlockState = self.collector.blockSignals(True)
            try:
                if inspector is None:
                    logger.warning(errMsg)
                    self._inspector = ErrorMsgInspector(self._collector, errMsg)
                    self.collector.clearAndSetComboBoxes([])
                else:
                    assert not errMsg, "Unexpected error message: {}".format(errMsg)
                    self._inspector = inspector
                    # Add and apply config values to the inspector
                    key = self.inspectorRegItem.identifier
                    cfg = self._inspectorStates.get(key, {})
                    logger.debug("Setting inspector settings from : {}".format(cfg))
                    self.inspector.config.unmarshall(cfg)
                    self._configTreeModel.setInvisibleRootItem(self.inspector.config)
                    self.configWidget.configTreeView.expandBranch()
                    self.collector.clearAndSetComboBoxes(self.inspector.axesNames())

                self._inspector.sigShowMessage.connect(self.sigShowMessage)
                self._inspector.sigUpdated.connect(self.testWalkDialog.setTestResult)
                self.wrapperLayout.addWidget(self.inspector)
            finally:
                self.collector.blockSignals(oldBlockState)
        finally:
            logger.debug("Enabling updates.")
            self.setUpdatesEnabled(True)

            self.updateWindowTitle()

            logger.debug("Emitting sigInspectorChanged({})".format(self.inspectorRegItem))
            self.sigInspectorChanged.emit(self.inspectorRegItem)


    def _storeInspectorState(self, inspectorRegItem, inspector):
        """ Store the settings values for the current inspector in a local dictionary.
            This dictionary is later used to store value for persistence.

            This function must be called after the inspector was drawn because that may update
            some derived config values (e.g. ranges)
        """
        if inspectorRegItem and inspector:
            key = inspectorRegItem.identifier
            logger.debug("_updateInspectorState: {} {}".format(key, type(inspector)))
            self._inspectorStates[key] = inspector.config.marshall()
        else:
            logger.debug("_updateInspectorState: no inspector")


    @QtSlot()
    def execPluginsDialog(self, label, registry):
        """ Shows the plugins dialog with the registered plugins
        """
        pluginsDialog = PluginsDialog(label, registry, parent=self)
        pluginsDialog.exec_()
        pluginsDialog.deleteLater()

        if pluginsDialog.result() == PluginsDialog.Accepted:
            logger.info("Accepted changes to {} registry.".format(label))

            logger.critical("Closing all windows and restarting eventloop")
            self.argosApplication.exit(EXIT_CODE_RESTART)
        else:
            logger.info("Cancelled changes to {} registry.".format(label))



    @QtSlot(str)
    def collectorContentsChanged(self, reason):
        """ Slot that updates the UI whenever the contents of the collector has changed.
        """
        logger.debug("collectorContentsChanged()")
        self.drawInspectorContents(reason=reason)


    @QtSlot(AbstractCti)
    def configContentsChanged(self, configTreeItem):
        """ Slot is called when an item has been changed by setData of the ConfigTreeModel.
            Will draw the window contents.
        """
        logger.debug("configContentsChanged: {}".format(configTreeItem))
        self.drawInspectorContents(reason=UpdateReason.CONFIG_CHANGED,
                                   origin=configTreeItem)

    def showMessage(self, msg):
        """ Shows a message to the user in the panel above the inspector
        """
        self.inspectorSelectionPane.showMessage(msg)


    def drawInspectorContents(self, reason, origin=None):
        """ Draws all contents of this window's inspector.
            The reason and origin parameters are passed on to the inspector's updateContents method.

            :param reason: string describing the reason for the redraw.
                Should preferably be one of the UpdateReason enumeration class, but new values may
                be used (which are then ignored by existing inspectors).
            :param origin: object with extra info on the reason
        """
        logger.debug("")
        logger.debug("-------- Drawing inspector of window: {} --------".format(self.windowTitle()))
        if PROFILING:
            self._profiler.enable()

        #self.showMessage(reason)
        self.showMessage('')  # clear message

        self.inspector.updateContents(reason=reason, initiator=origin)

        if PROFILING:
            self._profiler.disable()
        logger.debug("Finished draw inspector.\n")


    # TODO: to repotreemodel? Note that the functionality will be common to selectors.
    @QtSlot()
    def openFiles(self, fileNames=None, rtiRegItem=None, caption=None, fileMode=None):
        """ Lets the user select on or more files and opens it.

            :param fileNames: If None an open-file dialog allows the user to select files,
                otherwise the files are opened directly.
            :param rtiRegItem: Open the files as this type of registered RTI. None=autodetect.
            :param caption: Optional caption for the file dialog.
            :param fileMode: is passed to the file dialog.
            :rtype fileMode: QtWidgets.QFileDialog.FileMode constant
        """
        check_is_a_sequence(fileNames, allow_none=True)
        if fileNames is None:
            dialog = QtWidgets.QFileDialog(self, caption=caption)

            if rtiRegItem is None:
                nameFilter = 'All files (*);;' # Default show all files.
                nameFilter += self.argosApplication.rtiRegistry.getFileDialogFilter()
                if fileMode == QtWidgets.QFileDialog.Directory:
                    rtiRegItemName = 'Directory'
                else:
                    rtiRegItemName = ''
            else:
                nameFilter = rtiRegItem.getFileDialogFilter()
                nameFilter += ';;All files (*)'
                rtiRegItemName = rtiRegItem.name
            dialog.setNameFilter(nameFilter)

            if fileMode:
                dialog.setFileMode(fileMode)

            if dialog.exec_() == QtWidgets.QFileDialog.Accepted:
                fileNames = dialog.selectedFiles()
            else:
                fileNames = []

            # Only add files that were added via the dialog box (not via the command line).
            self._argosApplication.addToRecentFiles(fileNames, rtiRegItemName)

        fileRootIndex = None

        logger.debug("Adding files: {}".format(fileNames))
        for fileName in fileNames:
            fileRootIndex = self.argosApplication.repo.loadFile(fileName, rtiRegItem=rtiRegItem)

        if len(fileNames) == 1: # Only expand and open the file if the user selected one.
            logger.debug("Opening file: {}".format(fileNames[0]))
            self.repoWidget.repoTreeView.setExpanded(fileRootIndex, True)

        # Select last opened file
        if fileRootIndex is not None:
            self.repoWidget.repoTreeView.setCurrentIndex(fileRootIndex)


    def selectRtiByPath(self, path):
        """ Selects a repository tree item given a path, expanding nodes if along the way if needed.

            Returns (item, index) if the path was selected successfully, else raises an IndexError.
        """
        lastItem, lastIndex = self.repoWidget.repoTreeView.expandPath(path)
        self.repoWidget.repoTreeView.setCurrentIndex(lastIndex)
        return lastItem, lastIndex


    def trySelectRtiByPath(self, path):
        """ Selects a repository tree item given a path, expanding nodes if along the way if needed.

            Returns (item, index) if the path was selected successfully, else a warning is logged
            and (None, None) is returned.
        """
        try:
            logger.debug("Trying to select: {}".format(path))
            lastItem, lastIndex = self.repoWidget.repoTreeView.expandPath(path)
            self.repoWidget.repoTreeView.setCurrentIndex(lastIndex)
            self.repoWidget.repoTreeView.setFocus()
            return lastItem, lastIndex
        except Exception as ex:
            logger.warning("Unable to select {!r} because: {}".format(path, ex))
            if DEBUGGING:
                raise
            return None, None


    def openInWebBrowser(self, url):
        """ Opens url or file in an external documentation.

            Regular URLs are opened in the web browser, Local URLs are opened in the application
            that is used to open that type of file by default.
        """
        try:
            logger.debug("Opening URL: {}".format(url))
            qUrl = QUrl(url)
            QtGui.QDesktopServices.openUrl(qUrl)
        except Exception as ex:
            msg = "Unable to open URL {}. \n\nDetails: {}".format(url, ex)
            QtWidgets.QMessageBox.warning(self, "Warning", msg)
            logger.error(msg.replace('\n', ' '))


    def openInExternalApp(self, fileName):
        """ Opens url or file in an external documentation.

            Regular URLs are opened in the web browser, Local URLs are opened in the application
            that is used to open that type of file by default.
        """
        try:
            logger.debug("Opening URL: {}".format(fileName))
            if not os.path.exists(fileName):
                raise OSError("File doesn't exist.")
            url = QUrl.fromLocalFile(fileName)
            QtGui.QDesktopServices.openUrl(url)
        except Exception as ex:
            msg = "Unable to open file '{}'. \n\nDetails: {}".format(fileName, ex)
            QtWidgets.QMessageBox.warning(self, "Warning", msg)
            logger.error(msg.strip('\n'))


    def marshall(self):
        """ Returns a dictionary to save in the persistent settings
        """
        self._storeInspectorState(self.inspectorRegItem, self.inspector)

        twLayoutCfg, twCfg = self.testWalkDialog.marshall()

        layoutCfg = dict(
            repoWidget = self.repoWidget.marshall(),
            configTreeHeaders =  self.configWidget.configTreeView.marshall(),
            collectorHeaders = self.collector.tree.marshall(),
            winGeom = base64.b64encode(getWidgetGeom(self)).decode('ascii'),
            winState = base64.b64encode(getWidgetState(self)).decode('ascii'),
            testWalkDialog = twLayoutCfg,
        )

        cfg = dict(
            configWidget = self.configWidget.marshall(),
            curInspector = self.inspectorRegItem.identifier if self.inspectorRegItem else '',
            inspectors = self._inspectorStates,
            testWalkDialog = twCfg,
            layout = layoutCfg,
        )
        return cfg


    def unmarshall(self, cfg):
        """ Initializes itself from a config dict form the persistent settings.
        """
        self.configWidget.unmarshall(cfg.get('configWidget', {}))

        self._inspectorStates = cfg.get('inspectors', {})

        curInspector = cfg.get('curInspector')
        if curInspector:
            try:
                logger.debug("Setting inspector to: {}".format(curInspector))
                self.setInspectorById(curInspector)
            except KeyError as ex:
                logger.warning("No inspector with ID {!r}.: {}".format(curInspector, ex))

        layoutCfg = cfg.get('layout', {})

        self.repoWidget.unmarshall(layoutCfg.get('repoWidget', {}))
        self.configWidget.configTreeView.unmarshall(layoutCfg.get('configTreeHeaders', ''))
        self.collector.tree.unmarshall(layoutCfg.get('collectorHeaders', ''))

        self.testWalkDialog.unmarshall(layoutCfg.get('testWalkDialog', {}),
                                       cfg.get('testWalkDialog', {}))

        if 'winGeom' in layoutCfg:
            self.restoreGeometry(base64.b64decode(layoutCfg['winGeom']))
        if 'winState' in layoutCfg:
            self.restoreState(base64.b64decode(layoutCfg['winState']))



    @QtSlot()
    def cloneWindow(self):
        """ Opens a new window with the same inspector as the current window.
        """
        newWindow = self.argosApplication.addNewMainWindow(
            cfg=self.marshall(), inspectorFullName=self.inspectorRegItem.name)

        # Select the current item in the new window.
        currentItem, _currentIndex = self.repoWidget.repoTreeView.getCurrentItem()
        if currentItem:
            newWindow.trySelectRtiByPath(currentItem.nodePath)

        # Move the new window 24 pixels to the bottom right and raise it to the front.
        newGeomRect = newWindow.geometry()
        logger.debug("newGeomRect: x={}".format(newGeomRect.x()))
        newGeomRect.moveTo(newGeomRect.x() + 24, newGeomRect.y() + 24)

        newWindow.setGeometry(newGeomRect)
        logger.debug("newGeomRect: x={}".format(newGeomRect.x()))

        newWindow.raise_()


    @QtSlot()
    def activateAndRaise(self):
        """ Activates and raises the window.
        """
        logger.debug("Activate and raising window: {}".format(self.windowNumber))
        self.activateWindow()
        self.raise_()


    def event(self, ev):
        """ Detects the WindowActivate event. Pass all event through to the super class.
        """
        if ev.type() == QtCore.QEvent.WindowActivate:
            logger.debug("Window activated: {}".format(self.windowNumber))
            self.activateWindowAction.setChecked(True)

        return super(MainWindow, self).event(ev);


    def closeEvent(self, event):
        """ Called when closing this window.
        """
        logger.debug("closeEvent")

        # Save settings must be called here, at the point that there is still a windows open.
        # We can't use the QApplication.aboutToQuit signal because at that point the windows have
        # been closed
        self.argosApplication.saveSettingsIfLastWindow()
        self.finalize()
        self.argosApplication.removeMainWindow(self)
        event.accept()
        logger.debug("closeEvent accepted")


    @QtSlot()
    def about(self):
        """ Shows the about message window.
        """
        aboutDialog = AboutDialog(parent=self)
        aboutDialog.show()
        aboutDialog.addDependencyInfo()


    def addTestData(self):
        """ Adds test data to the repository
        """
        logger.info("Adding test data to the repository.")
        self.argosApplication.repo.insertItem(createArgosTestData())


    def showTestWalkDialog(self):
        """ Shows the test-walk dialog box
        """
        self.testWalkDialog.show()
        self.testWalkDialog.raise_()


    @QtSlot()
    def myTest(self):
        """ Function for small ad-hoc tests that can be called from the menu.
        """
        logger.info("--------- myTest function called --------------------")

        logger.info("--------- myTest function done --------------------")





# -*- coding: utf-8 -*-
# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Defines a global RTI registry to register repository tree item plugins.
"""
import logging
import os.path

from fnmatch import fnmatch

from argos.info import DEBUGGING
from argos.reg.basereg import BaseRegItem, BaseRegistry, RegType
from argos.repo.iconfactory import RtiIconFactory
from argos.utils.cls import check_class, is_a_color_str

logger = logging.getLogger(__name__)

ICON_COLOR_UNDEF = RtiIconFactory.COLOR_UNDEF
ICON_COLOR_UNKNOWN = RtiIconFactory.COLOR_UNKNOWN
ICON_COLOR_ERROR = RtiIconFactory.COLOR_ERROR
ICON_COLOR_MEMORY = RtiIconFactory.COLOR_MEMORY

ICON_COLOR_EXDIR = '#00BBFF'
ICON_COLOR_H5PY = '#00EE88'
ICON_COLOR_NCDF4 = '#0088FF'
ICON_COLOR_NUMPY = '#987456'
ICON_COLOR_PANDAS = '#FB9A99'
ICON_COLOR_PILLOW = '#FF40FF'
ICON_COLOR_SCIPY = ICON_COLOR_NUMPY
ICON_COLOR_JSON = '#880088'



class RtiRegItem(BaseRegItem):
    """ Class to keep track of a registered Repo Tree Item.
    """
    FIELDS  = BaseRegItem.FIELDS[:1] + ['iconColor', 'globs'] + BaseRegItem.FIELDS[1:]
    TYPES   = BaseRegItem.TYPES[:1] + [RegType.ColorStr, RegType.String] + BaseRegItem.TYPES[1:]
    LABELS  = BaseRegItem.LABELS[:1] + ['Icon Color', 'Globs'] + BaseRegItem.LABELS[1:]
    STRETCH = BaseRegItem.STRETCH[:1] + [False, True] + BaseRegItem.STRETCH[1:]

    COL_DECORATION = 0  # Display Icon in the main column

    def __init__(self, name='', absClassName='', pythonPath='', iconColor=ICON_COLOR_UNDEF, globs=''):
        """ Constructor. See the ClassRegItem class doc string for the parameter help.
        """
        super(RtiRegItem, self).__init__(name=name, absClassName=absClassName, pythonPath=pythonPath)
        check_class(globs, str)
        assert is_a_color_str(iconColor), \
            "Icon color for {} is not a color string: {!r}".format(self, iconColor)

        self._data['iconColor'] = iconColor
        self._data['globs'] = globs

    def __str__(self):
        return "<RtiRegItem: {}>".format(self.name)

    @property
    def iconColor(self):
        """ Icon color hex string.
        """
        return self._data['iconColor']


    @property
    def globList(self):
        """ Returns list of globs by splitting the globs string at the colons (:).
        """
        return self._data['globs'].split(';')


    def pathNameMatchesGlobs(self, path):
        """ Returns True if the file path matches one of the globs

            Matching is case-insensitive. See the Python fnmatch module for further info.
        """
        for glob in self.globList:
            if DEBUGGING:
                logger.debug("  glob '{}' -> match = {}".format(glob, fnmatch(path, glob)))
            if fnmatch(path, glob):
                return True

        return False


    def getFileDialogFilter(self):
        """ Returns a filters that can be used to construct file dialogs filters,
            for example: 'Text File (*.txt;*.text)'
        """
        # Remove any path info from the glob. E.g. '/mypath/prefix*1.nc' becomes '*.nc'
        extensions = ['*' + os.path.splitext(glob)[1] for glob in self.globList]
        return '{} ({})'.format(self.name, ';'.join(extensions))


    @property
    def decoration(self):
        """ The displayed icon.
        """
        rtiIconFactory = RtiIconFactory.singleton()

        if self._exception:
            return rtiIconFactory.getIcon(
                rtiIconFactory.ERROR, isOpen=False, color=rtiIconFactory.COLOR_ERROR)
        else:
            if self._cls is None:
                return rtiIconFactory.getIcon(
                    rtiIconFactory.ERROR, isOpen=False, color=rtiIconFactory.COLOR_UNKNOWN)
            else:
                return rtiIconFactory.getIcon(
                    self.cls._defaultIconGlyph, isOpen=False, color=self.iconColor)





class RtiRegistry(BaseRegistry):
    """ Class that can be used to register repository tree items (RTIs).

        Maintains a name to RtiClass mapping and an extension to RtiClass mapping.
        The extension in the extensionToRti assure that a unique RTI is used as default mapping,
        the extensions in the RtiRegItem class do not have to be unique and are used in the
        filter in the getFileDialogFilter function.
    """
    ITEM_CLASS = RtiRegItem

    DIRECTORY_REG_ITEM = RtiRegItem('Directory', 'argos.repo.filesytemrtis.DirectoryRti',
                                    iconColor=ICON_COLOR_UNKNOWN)

    def __init__(self):
        """ Constructor
        """
        super(RtiRegistry, self).__init__()
        self._extensionMap = {}


    def getRtiRegItemByExtension(self, filePath):
        """ Returns the first RtiRegItem class where filePath matches one of the globs patherns.
            Returns None if no class registered for the extension.
        """
        # Current implementation just returns the first rtiRegItem that contains the extension.
        logger.debug("{} getRtiRegItemByExtension, filePath: {}".format(self, filePath))
        for rtiRegItem in self._items:
            if rtiRegItem.pathNameMatchesGlobs(filePath):
                return rtiRegItem

        return None


    def getFileDialogFilter(self):
        """ Returns a filter that can be used in open file dialogs,
            for example: 'All files (*);;Txt (*.txt;*.text);;netCDF(*.nc;*.nc4)'
        """
        filters = []
        for regRti in self.items:
            filters.append(regRti.getFileDialogFilter())
        return ';;'.join(filters)


    def extraItemsForOpenAsMenu(self):
        """ Creates list of RtiRegItem to append to the 'open-as' and 'reload-as menus
        """
        # Add directory to the context menu so a an Exdir 'file' can be re-opened as a directory
        return [self.DIRECTORY_REG_ITEM]


    def getDefaultItems(self):
        """ Returns a list with the default plugins in the repo tree item registry.
        """

        # Note that when finding a plugin by extension, Argos uses the first one that matches.
        # Therefore put the defaults at the top of the list. The user can changed the order in the
        # plugin configuration dialog.

        hdfGlobs = '*.hdf5;*.h5;*.h5e;*.he5' # hdf extension is for HDF-4
        return [
            RtiRegItem('HDF-5 file',
                       'argos.repo.rtiplugins.hdf5.H5pyFileRti',
                       iconColor=ICON_COLOR_H5PY,
                       globs=hdfGlobs),

            RtiRegItem('Exdir file',
                       'argos.repo.rtiplugins.exdir.ExdirFileRti',
                       iconColor=ICON_COLOR_EXDIR,
                       globs='*.exdir'),

            RtiRegItem('NetCDF file',
                       'argos.repo.rtiplugins.ncdf.NcdfFileRti',
                       iconColor=ICON_COLOR_NCDF4,
                       globs='*.nc;*.nc4'),

            RtiRegItem('Pandas HDF file',
                       'argos.repo.rtiplugins.pandasio.PandasHdfFileRti',
                       iconColor=ICON_COLOR_PANDAS,
                       globs=hdfGlobs),

            RtiRegItem('Pandas CSV file',
                       'argos.repo.rtiplugins.pandasio.PandasCsvFileRti',
                       iconColor=ICON_COLOR_PANDAS,
                       globs='*.csv'),

            RtiRegItem('NumPy binary file',
                       'argos.repo.rtiplugins.numpyio.NumpyBinaryFileRti',
                       iconColor=ICON_COLOR_NUMPY,
                       globs='*.npy'),

            RtiRegItem('NumPy compressed file',
                       'argos.repo.rtiplugins.numpyio.NumpyCompressedFileRti',
                       iconColor=ICON_COLOR_NUMPY,
                       globs='*.npz'),

            RtiRegItem('NumPy text file',
                       'argos.repo.rtiplugins.numpyio.NumpyTextFileRti',
                       iconColor=ICON_COLOR_NUMPY,
                       #globs=['*.txt;*.text'),
                       globs='*.dat'),

            RtiRegItem('IDL save file',
                       'argos.repo.rtiplugins.scipyio.IdlSaveFileRti',
                       iconColor=ICON_COLOR_SCIPY,
                       globs='*.sav'),

            RtiRegItem('MATLAB file',
                       'argos.repo.rtiplugins.scipyio.MatlabFileRti',
                       iconColor=ICON_COLOR_SCIPY,
                       globs='*.mat'),

            RtiRegItem('Wav file',
                       'argos.repo.rtiplugins.scipyio.WavFileRti',
                       iconColor=ICON_COLOR_SCIPY,
                       globs='*.wav'),

            RtiRegItem('Pillow image',
                       'argos.repo.rtiplugins.pillowio.PillowFileRti',
                       iconColor=ICON_COLOR_PILLOW,
                       globs='*.bmp;*.eps;*.im;*.gif;*.jpg;*.jpeg;*.msp;*.pcx;*.png;*.ppm;*.spi;'
                             '*.tif;*.tiff;*.xbm;*.xv'),

            RtiRegItem('JSON file',
                       'argos.repo.rtiplugins.jsonio.JsonFileRti',
                       iconColor=ICON_COLOR_JSON,
                       globs='*.json'),
        ]


# The RTI registry is implemented as a singleton. This is necessary because
# in DirectoryRti._fetchAllChildren we need access to the registry.
# TODO: think of an elegant way to access the ArgosApplication.registry from there.
# TODO: Make use of the lib.cls.Singleton mixin (or metaclass when Python 3)
def createGlobalRegistryFunction():
    """ Closure to create the RtiRegistry singleton
    """
    globReg = RtiRegistry()

    def accessGlobalRegistry():
        return globReg

    return accessGlobalRegistry

# This is actually a function definition, not a constant
#pylint: disable=C0103

globalRtiRegistry = createGlobalRegistryFunction()
globalRtiRegistry.__doc__ = "Function that returns the RtiRegistry singleton common to all windows"






# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Repository items (RTIs) for browsing the file system
"""
import logging, os
from argos.repo.baserti import BaseRti
from argos.repo.iconfactory import RtiIconFactory
from argos.repo.registry import globalRtiRegistry, ICON_COLOR_UNKNOWN

logger = logging.getLogger(__name__)


class UnknownFileRti(BaseRti):
    """ A repository tree item that represents a file of unknown type.
        The file is not opened.
    """
    _defaultIconGlyph = RtiIconFactory.FILE

    def __init__(self, nodeName='', iconColor=ICON_COLOR_UNKNOWN, fileName=''):
        """ Constructor
        """
        super(UnknownFileRti, self).__init__(
            nodeName=nodeName, iconColor=iconColor, fileName=fileName)

        self._checkFileExists()


    def hasChildren(self):
        """ Returns False. Leaf nodes never have children. """
        return False



class DirectoryRti(BaseRti):
    """ A directory in the repository data tree.
    """
    _defaultIconGlyph = RtiIconFactory.FOLDER
    _defaultIconColor = RtiIconFactory.COLOR_UNKNOWN

    def __init__(self, nodeName='', iconColor=ICON_COLOR_UNKNOWN, fileName=''):
        """ Constructor
        """
        super(DirectoryRti, self).__init__(
            nodeName=nodeName, iconColor=iconColor, fileName=fileName)
        self._checkFileExists() # TODO: check for directory?


    def _fetchAllChildren(self):
        """ Gets all sub directories and files within the current directory.
            Does not fetch hidden files.
        """
        childItems = []
        fileNames = sorted(os.listdir(self._fileName), key=lambda s: s.lower())
        absFileNames = [os.path.join(self._fileName, fn) for fn in fileNames]

        for fileName, absFileName in zip(fileNames, absFileNames):
            if not fileName.startswith('.'):
                childItem = createRtiFromFileName(absFileName)
                childItems.append(childItem)

        return childItems


def _detectRtiFromFileName(fileName):
    """ Determines the type of RepoTreeItem to use given a file or directory name.
        Uses a DirectoryRti for directories without a registered extension and an UnknownFileRti
        if the file extension doesn't match one of the registered RTI globs.

        Returns (cls, regItem) tuple. Both the cls ond the regItem can be None.
        If the file is a directory without a registered extension, (DirectoryRti, None) is returned.
        If the file extension is not in the registry, (UnknownFileRti, None) is returned.
        If the cls cannot be imported (None, regItem) returned. regItem.exception will be set.
        Otherwise (cls, regItem) will be returned.

         Note that directories can have an extension (e.g. extdir archives). So it is not enough to
         just test if a file is a directory.
    """
    #_, extension = os.path.splitext(os.path.normpath(fileName))
    fullPath = os.path.normpath(os.path.abspath(fileName))
    rtiRegItem = globalRtiRegistry().getRtiRegItemByExtension(fullPath)
    if rtiRegItem is None:
        if os.path.isdir(fileName):
            cls = DirectoryRti
        else:
            logger.debug("No file RTI registered for path: {}".format(fullPath))
            cls = UnknownFileRti
    else:
         cls = rtiRegItem.getClass(tryImport=True) # cls can be None

    return cls, rtiRegItem


def createRtiFromFileName(fileName):
    """ Determines the type of RepoTreeItem to use given a file or directory name and creates it.
        Uses a DirectoryRti for directories without registered extensions and an UnknownFileRti if the file
        extension doesn't match one of the registered RTI extensions.
    """
    cls, rtiRegItem = _detectRtiFromFileName(fileName)
    assert not (cls is None and rtiRegItem is None), "cls and rtiRegItem both none."

    iconColor = rtiRegItem.iconColor if rtiRegItem else ICON_COLOR_UNKNOWN

    if cls is None:
        logger.warning("Unable to import plugin {}: {}"
                       .format(rtiRegItem.name, rtiRegItem.exception))
        rti = UnknownFileRti.createFromFileName(fileName, ICON_COLOR_UNKNOWN)
        rti.setException(rtiRegItem.exception)
    else:
        logger.debug("Calling createFromFileName: {} ({}, {})".format(cls, fileName, iconColor))
        rti = cls.createFromFileName(fileName, iconColor)

    assert rti, "Sanity check failed (createRtiFromFileName). Please report this bug."

    return rti




# -*- coding: utf-8 -*-

# This file is part of Argos.
#
# Argos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Argos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Argos. If not, see <http://www.gnu.org/licenses/>.

""" Some simple Config Tree Items
"""
import logging, copy

from argos.config.abstractcti import AbstractCti, AbstractCtiEditor
from argos.qt import  Qt, QtCore, QtWidgets, QtSlot
from argos.qt.misc import setWidgetSizePolicy
from argos.utils.cls import check_class
from argos.utils.misc import NOT_SPECIFIED


logger = logging.getLogger(__name__)

# Use setIndexWidget()?



class ChoiceCti(AbstractCti):
    """ Config Tree Item to store a choice between strings.

        A QCombobox will pop-up if the user clicks on the cell to edit it.
    """
    def __init__(self, nodeName, defaultData=0, configValues=None, displayValues=None,
                 editable=False, insertPolicy=QtWidgets.QComboBox.InsertAtBottom,
                 completer=NOT_SPECIFIED):
        """ Constructor.

            The defaultData is an integers that is used to store the currentIndex.
            The displayValues parameter must be a list of strings, which will be displayed in the
            combo box. The configValues should be a list of the same size with the configValues
            that each 'choice' represents, e.g. choice 'dashed' maps to configValue Qt.DashLine.
            If displayValues is None, the configValues are used as displayValues.

            The defaultData can be set to a negative value, e.g. use -1 to select the last item
            by default. However, it will be converted to a positive value in the constructor.

            :param defaultData: the default index in the combobox that is used for editing
            :param configValues: list of configValue items to populate the combobox with
            :param displayValues: list of possible displayValue items.
                If None the config values are used as display values as well.
            :param editable: if True the combobox will be editable
            :param insertPolicy: determines where user-inserted items should appear in the combobox
                must be of type QtWidgets.QComboBox.InsertPolicy
            :param completer: a QCompleter that will be used by the combobox for auto-completion.
                If NotSpecified a default completer is used, i.e. case-insensitive auto-completion.
                Use None, 0 or False to disable auto-completion.

            For the  parameters see the AbstractCti constructor documentation.

        """
        check_class(insertPolicy, QtWidgets.QComboBox.InsertPolicy)
        if completer: # completer may be False
            check_class(completer, QtWidgets.QCompleter)

        self.editable = editable
        self.insertPolicy = insertPolicy
        self.completer = completer

        self._configValues = [] if configValues is None else configValues
        if displayValues is None:
            self._displayValues = copy.copy(self._configValues)
        else:
            assert not editable, "No separate displayValues may be set if the CTI is editable"
            self._displayValues = displayValues

        assert len(self._configValues) == len(self._displayValues),\
            "If set, configValues must have the same length as displayValues."

        self._defaultConfigValues = copy.copy(self._configValues)

        # Set after self._displayValues are defined. The parent constructor call _enforceDataType
        super(ChoiceCti, self).__init__(nodeName, defaultData)


    def _enforceDataType(self, data):
        """ Converts to int so that this CTI always stores that type.

            The data be set to a negative value, e.g. use -1 to select the last item
            by default. However, it will be converted to a positive by this method.
        """
        idx = int(data)
        if idx < 0:
            idx += len(self._displayValues)
        assert 0 <= idx < len(self._displayValues), \
            "Index should be >= 0 and < {}. Got {}".format(len(self._displayValues), idx)
        return idx


    @property
    def configValue(self):
        """ The currently selected configValue
        """
        return self._configValues[self.data]


    @property
    def iterConfigValues(self):
        """ Returns a generator expression to iterate over al config values
        """
        return (value for value in self._configValues)


    @property
    def iterDisplayValues(self):
        """ Returns a generator expression to iterate over al display values
        """
        return (value for value in self._displayValues)


    def _dataToString(self, data):
        """ Conversion function used to convert the (default)data to the display value.
        """
        choices = self._displayValues
        return str(choices[data])


    def _nodeMarshall(self):
        """ Returns the non-recursive marshalled value of this CTI. Is called by marshall()
        """
        if self.editable:
            return [self.data, self._configValues]
        else:
            return self.data


    def _nodeUnmarshall(self, data):
        """ Initializes itself non-recursively from data. Is called by unmarshall()
        """
        if self.editable:
            curIdx, configValues = data
            self._configValues = configValues
            self._displayValues = configValues
            self.data = curIdx
        else:
            self.data = data


    @property
    def debugInfo(self):
        """ Returns the string with debugging information
        """
        return repr(self._displayValues)


    def createEditor(self, delegate, parent, option):
        """ Creates a ChoiceCtiEditor.
            For the parameters see the AbstractCti constructor documentation.
        """
        return ChoiceCtiEditor(self, delegate, parent=parent)


    def insertValue(self, pos, configValue, displayValue=None):
        """ Will insert the configValue in the configValues and the displayValue in the
            displayValues list.
            If displayValue is None, the configValue is set in the displayValues as well
        """
        self._configValues.insert(pos, configValue)
        self._displayValues.insert(pos, displayValue if displayValue is not None else configValue)


    def removeValue(self, configValue):
        """ Will remove the configValue and the corresponding displayValue from the lists.
        """
        self.removeValueByIndex(self._configValues.index(configValue))


    def removeValueByIndex(self, index):
        """ Will remove the configValue and the corresponding displayValue given the index.
        """
        del self._configValues[index]
        del self._displayValues[index]


class ChoiceCtiEditor(AbstractCtiEditor):
    """ A CtiEditor which contains a QCombobox for editing ChoiceCti objects.
    """
    def __init__(self, cti, delegate, parent=None):
        """ See the AbstractCtiEditor for more info on the parameters
        """
        super(ChoiceCtiEditor, self).__init__(cti, delegate, parent=parent)

        comboBox = QtWidgets.QComboBox()
        comboBox.setEditable(cti.editable)
        comboBox.setInsertPolicy(cti.insertPolicy)
        setWidgetSizePolicy(comboBox, QtWidgets.QSizePolicy.Expanding, None)

        if cti.completer is not NOT_SPECIFIED:
            comboBox.setCompleter(cti.completer if cti.completer else None)

        comboBox.addItems(cti._displayValues)

        # Store the configValue in the combo box, although it's not currently used.
        for idx, configValue in enumerate(cti._configValues):
            comboBox.setItemData(idx, configValue, role=Qt.UserRole)

        comboBox.activated.connect(self.comboBoxActivated)
        comboBox.model().rowsInserted.connect(self.comboBoxRowsInserted)
        self.comboBox = self.addSubEditor(comboBox, isFocusProxy=True)

        # Install an eventFilter on the QListView that pops-up and contains all the items.
        # This allows the user to remove items by pressing the delete key in the list.
        self._comboboxListView = comboBox.view()
        self._comboboxListView.installEventFilter(self)


    def finalize(self):
        """ Is called when the editor is closed. Disconnect signals.
        """
        self._comboboxListView.removeEventFilter(self)
        self.comboBox.model().rowsInserted.disconnect(self.comboBoxRowsInserted)
        self.comboBox.activated.disconnect(self.comboBoxActivated)
        super(ChoiceCtiEditor, self).finalize()


    def setData(self, data):
        """ Provides the main editor widget with a data to manipulate.
        """
        self.comboBox.setCurrentIndex(data)


    def getData(self):
        """ Gets data from the editor widget.
        """
        return self.comboBox.currentIndex()


    @QtSlot(int)
    def comboBoxActivated(self, index):
        """ Is called when the user chooses an item in the combo box. The item's index is passed.
            Note that this signal is sent even when the choice is not changed.
        """
        self.delegate.commitData.emit(self)


    @QtSlot(QtCore.QModelIndex, int, int)
    def comboBoxRowsInserted(self, _parent, start, end):
        """ Called when the user has entered a new value in the combobox.
            Puts the combobox values back into the cti.
        """
        assert start == end, "Bug, please report: more than one row inserted"
        configValue = self.comboBox.itemText(start)
        logger.debug("Inserting {!r} at position {} in {}"
                     .format(configValue, start, self.cti.nodePath))
        self.cti.insertValue(start, configValue)


    def eventFilter(self, watchedObject, event):
        """ Deletes an item from an editable combobox when the delete or backspace key is pressed
            in the list of items, or when ctrl-delete or ctrl-back space is pressed in the
            line-edit.

            When the combobox is not editable the filter does nothing.
        """
        if self.comboBox.isEditable() and event.type() == QtCore.QEvent.KeyPress:
            key = event.key()
            if key in (Qt.Key_Delete, Qt.Key_Backspace):
                if (watchedObject == self._comboboxListView
                    or (watchedObject == self.comboBox
                        and event.modifiers() == Qt.ControlModifier)):

                    index = self._comboboxListView.currentIndex()
                    if index.isValid():
                        row = index.row()
                        logger.debug("Removing item {} from the combobox: {}"
                                     .format(row, self._comboboxListView.model().data(index)))
                        self.cti.removeValueByIndex(row)
                        self.comboBox.removeItem(row)
                        return True

        # Calling parent event filter, which may filter out other events.
        return super(ChoiceCtiEditor, self).eventFilter(watchedObject, event)

